/*
    Copyright 2006-2009 Nagoya University

    This file is part of NDD (Near Duplicate Detector).

    NDD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    NDD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NDD.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ndd/fvd.h>
#include <ndd/error.h>
#include <ndd/file.h>
#include <ndd/memory.h>

#include <getopt.h>

typedef struct __options
{
    const char **input_fvdfiles;
    ndd_uint32   input_fvdfile_count;

    const char  *output_fvdfile;

} options;

static const struct option long_options[] = {
    {"src", required_argument, 0, 0},
    {"dst", required_argument, 0, 0},
    {NULL, 0, 0, 0},
};

static ndd_result
show_usage()
{
    fputs("Usage: catfvd --src=<input-fvdfile> [--src=<input-fvdfile> ...]\n", stderr);
    fputs("              --dst=<outut-fvdfile>\n", stderr);
    return NDD_SUCCESS;
}

static ndd_result
get_options(options *o, int argc, char *argv[])
{
    NDD_START;

    NDD_WHILE(1){
	int index = -1;
	int c = getopt_long(argc, argv, "", long_options, &index);

	if(c == -1){
	    break;
	}
	switch(c){
	case '?':
	    show_usage();
	    break;

	default:
	    switch(index){
	    case 0:
		NDD_VERIFY_RESULT(ndd_realloc(&o->input_fvdfiles, sizeof(const char *) * (o->input_fvdfile_count + 1)));
		NDD_IFOK{
		    o->input_fvdfiles[o->input_fvdfile_count++] = optarg;
		}
		break;
	    case 1:
		o->output_fvdfile = optarg;
		break;
	    }
	}
    }

    /*  å  */
    if(o->input_fvdfile_count < 1){
	fputs("No source fvd filename was given.  Use --src=<FILENAME>.\n", stderr);
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }else
    if(o->output_fvdfile == NULL){
	fputs("No destination fvd filename was given.  Use --dst=<FILENAME>.\n", stderr);
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }

    NDD_RETURN;
}

int main(int argc, char *argv[])
{
    NDD_START;

    options opt = {NULL, 0, NULL};
    NDD_VERIFY_RESULT(get_options(&opt, argc, argv));

    ndd_fvd *output_fvd = NULL;
    NDD_VERIFY_RESULT_OR(ndd_fvd_create(&output_fvd)){
	fputs("Failed to create fvd object.\n", stderr);
    }

    NDD_VERIFY_RESULT_OR(ndd_fvd_assign_file(output_fvd,
					     opt.output_fvdfile, 
					     ndd_file_mode_write)){
	fputs("Failed to create fvd object.\n", stderr);
    }

    ndd_uint32 i;
    NDD_FOR(i = 0; i < opt.input_fvdfile_count; i++){
	ndd_fvd *input_fvd = NULL;
	NDD_VERIFY_RESULT_OR(ndd_fvd_create(&input_fvd)){
	    fputs("Failed to create fvd object.\n", stderr);
	}

//	FILE *input_fp = NULL;
//	NDD_VERIFY((input_fp = fopen(opt.input_fvdfiles[i], "r")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
//	NDD_VERIFY_RESULT_OR(ndd_fvd_read_from_file(input_fvd, input_fp)){
//	    fputs("Failed to read fvd from file.\n", stderr);
//	}
	NDD_VERIFY_RESULT_OR(ndd_fvd_assign_file(input_fvd, opt.input_fvdfiles[i], ndd_file_mode_read)){
	    fputs("Failed to read fvd from file.\n", stderr);
	}
//	NDD_VERIFY_RESULT(ndd_fvd_concatenate(output_fvd, output_fvd, input_fvd));
	NDD_VERIFY_RESULT(ndd_fvd_append(output_fvd, input_fvd));

//	if(input_fp != NULL){
//	    fclose(input_fp);
//	    input_fp = NULL;
//	}
	if(input_fvd != NULL){
	    ndd_fvd_destroy(&input_fvd);
	}
    }

//    FILE *output_fp = NULL;
//    NDD_VERIFY((output_fp = fopen(opt.output_fvdfile, "w")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
//    NDD_VERIFY_RESULT_OR(ndd_fvd_write_to_file(output_fvd, output_fp)){
//	fputs("Failed to write fvd to file.\n", stderr);
//    }
//    if(output_fp != NULL){
//	fclose(output_fp);
//	output_fp = NULL;
//    }

    if(output_fvd != NULL){
	ndd_fvd_destroy(&output_fvd);
    }
    
    NDD_IFOK{
	return EXIT_SUCCESS;
    }else{
	fprintf(stderr, "Error code = %lu\n", __return_code);
	return EXIT_FAILURE;
    }
}

/*  end of file  */
