/*
    Copyright 2006-2009 Nagoya University

    This file is part of NDD (Near Duplicate Detector).

    NDD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    NDD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NDD.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ndd/sfd.h>
#include <ndd/error.h>
#include <ndd/file.h>
#include <ndd/memory.h>

#include <getopt.h>

typedef struct __options
{
    const char **input_sfdfiles;
    ndd_uint32   input_sfdfile_count;

    const char  *output_sfdfile;

} options;

static const struct option long_options[] = {
    {"src", required_argument, 0, 0},
    {"dst", required_argument, 0, 0},
    {NULL, 0, 0, 0},
};

static ndd_result
show_usage()
{
    fputs("Usage: catsfd --src=<input-sfdfile> [--src=<input-sfdfile> ...]\n", stderr);
    fputs("              --dst=<outut-sfdfile>\n", stderr);
    return NDD_SUCCESS;
}

static ndd_result
get_options(options *o, int argc, char *argv[])
{
    NDD_START;

    NDD_WHILE(1){
	int index = -1;
	int c = getopt_long(argc, argv, "", long_options, &index);

	if(c == -1){
	    break;
	}
	switch(c){
	case '?':
	    show_usage();
	    break;

	default:
	    switch(index){
	    case 0:
		NDD_VERIFY_RESULT(ndd_realloc(&o->input_sfdfiles, sizeof(const char *) * (o->input_sfdfile_count + 1)));
		NDD_IFOK{
		    o->input_sfdfiles[o->input_sfdfile_count++] = optarg;
		}
		break;
	    case 1:
		o->output_sfdfile = optarg;
		break;
	    }
	}
    }

    /*  å  */
    if(o->input_sfdfile_count < 1){
	fputs("No source sfd filename was given.  Use --src=<FILENAME>.\n", stderr);
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }else
    if(o->output_sfdfile == NULL){
	fputs("No destination sfd filename was given.  Use --dst=<FILENAME>.\n", stderr);
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }

    NDD_RETURN;
}

int main(int argc, char *argv[])
{
    NDD_START;

    options opt = {NULL, 0, NULL};
    NDD_VERIFY_RESULT(get_options(&opt, argc, argv));

    ndd_sfd *output_sfd = NULL;
    NDD_VERIFY_RESULT_OR(ndd_sfd_create(&output_sfd)){
	fputs("Failed to create sfd object.\n", stderr);
    }

    ndd_uint32 i;
    NDD_FOR(i = 0; i < opt.input_sfdfile_count; i++){
	ndd_sfd *input_sfd = NULL;
	NDD_VERIFY_RESULT_OR(ndd_sfd_create(&input_sfd)){
	    fputs("Failed to create sfd object.\n", stderr);
	}

	FILE *input_fp = NULL;
	NDD_VERIFY((input_fp = fopen(opt.input_sfdfiles[i], "r")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
	NDD_VERIFY_RESULT_OR(ndd_sfd_read_from_file(input_sfd, input_fp)){
	    fputs("Failed to read sfd from file.\n", stderr);
	}
//	NDD_VERIFY_RESULT_OR(ndd_sfd_assign_file(sfd, opt.input_sfdfile, ndd_file_mode_read)){
//	    fputs("Failed to read sfd from file.\n", stderr);
//	}
	NDD_VERIFY_RESULT(ndd_sfd_concatenate(output_sfd, output_sfd, input_sfd));

	if(input_fp != NULL){
	    fclose(input_fp);
	    input_fp = NULL;
	}
	if(input_sfd != NULL){
	    ndd_sfd_destroy(&input_sfd);
	}
    }

    FILE *output_fp = NULL;
    NDD_VERIFY((output_fp = fopen(opt.output_sfdfile, "w")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
    NDD_VERIFY_RESULT_OR(ndd_sfd_write_to_file(output_sfd, output_fp)){
	fputs("Failed to write sfd to file.\n", stderr);
    }
    if(output_fp != NULL){
	fclose(output_fp);
	output_fp = NULL;
    }

    if(output_sfd != NULL){
	ndd_sfd_destroy(&output_sfd);
    }
    
    NDD_IFOK{
	return EXIT_SUCCESS;
    }else{
	fprintf(stderr, "Error code = %lu\n", __return_code);
	return EXIT_FAILURE;
    }
}

/*  end of file  */
