/*
    Copyright 2006-2009 Nagoya University

    This file is part of NDD (Near Duplicate Detector).

    NDD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    NDD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NDD.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ndd/sfd.h>
#include <ndd/file.h>
#include <ndd/error.h>
#include <ndd/defaults.h>
#include <getopt.h>

typedef struct __options
{
    const char *input_sfdfile;   /*  Ϥ FVD ե̾  */
    const char *output_sfdfile;  /*  Ϥ FVD ե̾  */
    
    ndd_uint32  frame_count;     /*  ɬ  */
    ndd_uint32  frame_offset;    /*  ǥե = 0  */
    ndd_uint32  repeat_count;    /*  ǥե = 1  */

} options;

static const struct option long_options[]={
    {"src",          required_argument, 0, 0},
    {"dst",          required_argument, 0, 0},
    {"frame-count",  required_argument, 0, 0},
    {"frame-offset", required_argument, 0, 0},
    {"repeat",       optional_argument, 0, 0},
    {NULL, 0, 0, 0},
};

static ndd_result
show_usage()
{
    fputs("Usage: cutsfd --src=<input-sfdfile> --dst=<output-sfdfile>\n", stderr);
    fputs("              --frame-count=<frame-count>\n", stderr);
    fputs("             [--frame-offset=<frame-offset>]\n", stderr);
    fputs("              --repeat[=<repeat-count>]\n", stderr);
    return NDD_SUCCESS;
}

static ndd_result
get_options(options *o, int argc, char *argv[])
{
    NDD_START;

    ndd_bool input_sfdfile_isset      = false;
    ndd_bool output_sfdfile_isset     = false;
    ndd_bool frame_count_isset	      = false;
    ndd_bool frame_offset_isset	      = false;
    ndd_bool repeat_count_isset       = false;

    ndd_bool valid = true;
    NDD_WHILE(1){
	int index = -1;
	int c = getopt_long(argc, argv, "", long_options, &index);

	if(c == -1){
	    break;
	}
	switch(c){
	case '?':
	    show_usage();
	    valid = false;
	    break;

	default:
	    switch(index){
	    case 0:
		o->input_sfdfile = optarg;
		input_sfdfile_isset = true;
		break;
	    case 1:
		o->output_sfdfile = optarg;
		output_sfdfile_isset = true;
		break;
	    case 2:
		o->frame_count = atol(optarg);
		frame_count_isset = true;
		break;
	    case 3:
		o->frame_offset = atol(optarg);
		frame_offset_isset = true;
		break;
	    case 4:
	    	if(optarg == NULL){
		    o->repeat_count = 0;
	    	}else{
		    o->repeat_count = atol(optarg);
	    	}
		repeat_count_isset = true;
		break;
	    }
	}
    }

    /*  å  */
    if(valid){
	if(!input_sfdfile_isset){
	    fputs("No input-sfd filename was given.  Use --src=<input-sfdfile>.\n", stderr);
	    valid = false;
	}
	if(!output_sfdfile_isset){
	    fputs("No output-sfd filename was given.  Use --dst=<output-sfdfile>.\n", stderr);
	    valid = false;
	}
	if(!frame_count_isset){
	    fputs("Frame-count was not given.  Use --frame-count=<frame-count>.\n", stderr);
	    valid = false;
	}
	if(!frame_offset_isset){
	    fprintf(stdout, "Frame-offset was not given.  Using default offset %lu.\n", o->frame_offset);
	}
	if(!repeat_count_isset){
	    fprintf(stdout, "Repeat-count was not given.  Using default repeat-count %lu.\n", o->repeat_count);
	}
    }
    if(!valid){
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }

    NDD_RETURN;	
}

int main(int argc, char *argv[])
{
    NDD_START;

    options opt = {NULL, NULL, 0, 0, 1};
    NDD_VERIFY_RESULT(get_options(&opt, argc, argv));
    
    ndd_sfd *input_sfd = NULL;
    NDD_VERIFY_RESULT_OR(ndd_sfd_create(&input_sfd)){
	fputs("Failed to create sfd object.\n", stderr);
    }

    FILE *input_fp = NULL;
    NDD_VERIFY((input_fp = fopen(opt.input_sfdfile, "r")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
    NDD_VERIFY_RESULT_OR(ndd_sfd_read_from_file(input_sfd, input_fp)){
	fputs("Failed to read sfd from file.\n", stderr);
    }
    if(input_fp != NULL){
	fclose(input_fp);
	input_fp = NULL;
    }

    ndd_uint32 i;
    NDD_FOR(i = 0; opt.repeat_count == 0 || i < opt.repeat_count; i++){

	char filename[FILENAME_MAX];
	if(opt.repeat_count == 0 &&
	   opt.frame_offset + (i + 1) * opt.frame_count >= input_sfd->header->frame_offset + input_sfd->header->frame_count){
	    break;
	}
	if(opt.repeat_count == 1){
	    strncpy(filename, opt.output_sfdfile, FILENAME_MAX);
	}else{
	    snprintf(filename, FILENAME_MAX, "%s.%03lu", opt.output_sfdfile, i + 1);
	}

	ndd_sfd *output_sfd = NULL;
	NDD_VERIFY_RESULT_OR(ndd_sfd_create(&output_sfd)){
	    fputs("Failed to create sfd object.\n", stderr);
	}

	NDD_VERIFY_RESULT(ndd_sfd_cut(output_sfd, input_sfd, opt.frame_offset + i * opt.frame_count, opt.frame_count));

	FILE *output_fp = NULL;
	NDD_VERIFY((output_fp = fopen(filename, "w")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
	NDD_VERIFY_RESULT_OR(ndd_sfd_write_to_file(output_sfd, output_fp)){
	    fputs("Failed to write sfd to file.\n", stderr);
	}
	if(output_fp != NULL){
	    fclose(output_fp);
	    output_fp = NULL;
	}
	if(output_sfd != NULL){
	    ndd_sfd_destroy(&output_sfd);
	}
    }

    if(input_sfd != NULL){
	ndd_sfd_destroy(&input_sfd);
    }
    
    NDD_IFOK{
	return EXIT_SUCCESS;
    }else{
	fprintf(stderr, "Error code = %lu\n", __return_code);
	return EXIT_FAILURE;
    }
}

/*  end of file  */
