/*
    Copyright 2006-2009 Nagoya University

    This file is part of NDD (Near Duplicate Detector).

    NDD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    NDD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NDD.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ndd/mkffd.h>
#include <ndd/ffd.h>
#include <ndd/ffs.h>
#include <ndd/fvd.h>
#include <ndd/error.h>
#include <ndd/file.h>
#include <ndd/defaults.h>

#include <getopt.h>
typedef struct __options
{
    const char *input_fvdfile;   /*  Ϥ FVD ե̾  */
    const char *input_ffsfile;   /*  Ϥ FFS ե̾  */
    const char *output_ffdfile;  /*  Ϥ FFD ե̾  */

} options;

static const struct option long_options[]={
    {"fvd", required_argument, 0, 0},
    {"ffs", required_argument, 0, 0},
    {"ffd", required_argument, 0, 0},
    {NULL, 0, 0, 0},
};


static ndd_result
show_usage()
{
    fputs("Usage: mkffd --fvd=<input-fvdfile> --ffs=<input-ffsfile> --ffd=<output-ffdfile>\n", stderr);
    return NDD_SUCCESS;
}   

static ndd_result
get_options(options *o, int argc, char *argv[])
{
    NDD_START;

    ndd_bool input_fvdfile_isset  = false;
    ndd_bool input_ffsfile_isset  = false;
    ndd_bool output_ffdfile_isset = false;
 
    ndd_bool valid = true;
    NDD_WHILE(1){
	int index = -1;
	int c = getopt_long(argc, argv, "", long_options, &index);

	if(c == -1){
	    break;
	}
	switch(c){
	case '?':
	    show_usage();
	    valid = false;
	    break;

	default:
	    switch(index){	
	    case 0:
		o->input_fvdfile = optarg;
		input_fvdfile_isset = true;
		break;
	    case 1:
		o->input_ffsfile = optarg;
		input_ffsfile_isset = true;
		break;
	    case 2:
		o->output_ffdfile = optarg;
		output_ffdfile_isset = true;
		break;
	    }
	}
    }

    /*  å  */
    if(valid){
	if(o->input_fvdfile == NULL){
	    fputs("No fvd filename was given.  Use --fvd=<FILENAME>.\n", stderr);
	    NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
	}
	if(o->input_ffsfile == NULL){
	    fputs("No ffs filename was given.  Use --ffs=<FILENAME>.\n", stderr);
	    NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
	}
	if(o->output_ffdfile == NULL){
	    fputs("No ffd filename was given.  Use --ffd=<FILENAME>.\n", stderr);
	    NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
	}
    }
    if(!valid){
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }
    
    NDD_RETURN;
}

int main(int argc, char *argv[])
{
    NDD_START;

    options opt = {NULL, NULL, NULL};
    NDD_VERIFY_RESULT(get_options(&opt, argc, argv));	

    ndd_fvd *fvd = NULL;
    NDD_VERIFY_RESULT_OR(ndd_fvd_create(&fvd)){
	fputs("Failed to create fvd object.\n", stderr);
    }
    ndd_ffs *ffs = NULL;
    NDD_VERIFY_RESULT_OR(ndd_ffs_create(&ffs)){
	fputs("Failed to create ffs object.\n", stderr);
    }
    ndd_ffd *ffd = NULL;
    NDD_VERIFY_RESULT_OR(ndd_ffd_create(&ffd)){
	fputs("Failed to create ffd object.\n", stderr);
    }

    FILE *input_fp = NULL;
//    NDD_VERIFY((input_fp = fopen(opt.input_fvdfile, "r")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
//    NDD_VERIFY_RESULT_OR(ndd_fvd_read_from_file(fvd, input_fp)){
    NDD_VERIFY_RESULT_OR(ndd_fvd_assign_file(fvd, opt.input_fvdfile, ndd_file_mode_read)){
	fputs("Failed to read fvd from file.\n", stderr);
    }
//    if(input_fp != NULL){
//	fclose(input_fp);
//	input_fp = NULL;
//    }

    NDD_VERIFY((input_fp = fopen(opt.input_ffsfile, "r")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
    NDD_VERIFY_RESULT_OR(ndd_ffs_read_from_file(ffs, input_fp)){
	fputs("Failed to read ffs from file.\n", stderr);
    }
    if(input_fp != NULL){
	fclose(input_fp);
	input_fp = NULL;
    }

    NDD_VERIFY_RESULT_OR(ndd_create_ffd_from_fvd(ffd, fvd, ffs, ndd_callback_default_handler, NULL)){
	fputs("Failed to create ffd from fvd.\n", stderr);
    }
     
    FILE *output_fp = NULL;
    NDD_VERIFY((output_fp = fopen(opt.output_ffdfile, "w")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
    NDD_VERIFY_RESULT_OR(ndd_ffd_write_to_file(ffd, output_fp)){
	fputs("Failed to write ffd into file.\n", stderr);
    }

    if(ffd != NULL){
	ndd_ffd_destroy(&ffd);
    }
    if(ffs != NULL){
	ndd_ffs_destroy(&ffs);
    }
    if(fvd != NULL){
	ndd_fvd_destroy(&fvd);
    }
    if(output_fp != NULL){
	fclose(output_fp);
    }
    
    NDD_IFOK{
	return EXIT_SUCCESS;
    }else{
	fprintf(stderr, "Error code = %lu\n", __return_code);
	return EXIT_FAILURE;
    }
}

/*  end of file  */
