/*
    Copyright 2006-2009 Nagoya University

    This file is part of NDD (Near Duplicate Detector).

    NDD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    NDD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NDD.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ndd/mkffs.h>
#include <ndd/ffs.h>
#include <ndd/fvd.h>
#include <ndd/error.h>
#include <ndd/file.h>
#include <ndd/defaults.h>

#include <getopt.h>

typedef struct __options
{
    const char *input_fvdfile;
    const char *output_ffsfile;
	
    ndd_float8 frame_proportion;
    ndd_uint32 frame_dimension;
    ndd_uint32 frame_sample_count;

} options;

static const struct option long_options[]={
    {"fvd", required_argument, 0, 0},
    {"ffs", required_argument, 0, 0},
    {"proportion", required_argument, 0, 0},
    {"dimension", required_argument, 0, 0},
    {"sample-count", required_argument, 0, 0},
    {NULL, 0, 0, 0},
};


static ndd_result
show_usage()
{
    fputs("Usage: mkffs --fvd=<input-fvdfile> --ffs=<output-ffsfile>.\n", stderr);
    fputs("             {--proportion=<number> | --dimension=<number>}\n", stderr);
    fputs("             [--sample-count=<number>]\n", stderr);
    return NDD_SUCCESS;
}

static ndd_result
get_options(options *o, int argc, char *argv[])
{
    NDD_START;

    ndd_bool input_fvdfile_isset      = false;
    ndd_bool output_ffsfile_isset     = false;
    ndd_bool frame_proportion_isset   = false;
    ndd_bool frame_dimension_isset    = false;
    ndd_bool frame_sample_count_isset = false;

    ndd_bool valid = true;
    NDD_WHILE(1){
	int index = -1;
	int c = getopt_long(argc, argv, "", long_options, &index);

	if(c == -1){
	    break;
	}
	switch(c){
	case '?':
	    show_usage();
	    valid = false;
	    break;

	default:
	    switch(index){
	    case 0:
		o->input_fvdfile = optarg;
		input_fvdfile_isset = true;
		break;
	    case 1:
		o->output_ffsfile = optarg;
		output_ffsfile_isset = true;
		break;
	    case 2:
		o->frame_proportion = atof(optarg);
		frame_proportion_isset = true;
		break;
	    case 3:
		o->frame_dimension = atol(optarg);
		frame_dimension_isset = true;
		break;
	    case 4:
		o->frame_sample_count = atol(optarg);
		frame_sample_count_isset = true;
		break;
	    }
	}
    }

    /*  å  */
    if(valid){
	if(!input_fvdfile_isset){
	    fputs("No fvd filename was given.  Use --fvd=<FILENAME>.\n", stderr);
	    valid = false;
	}
	if(!output_ffsfile_isset){
	    fputs("No ffs filename was given.  Use --ffs=<FILENAME>.\n", stderr);
	    valid = false;
	}
	if(!frame_dimension_isset && !frame_proportion_isset){
	    fprintf(stdout, "Dimension nor proportion was given.  Using default dimension size %d.\n", DEFAULT_FRAME_DIMENSION);
	    o->frame_dimension  = DEFAULT_FRAME_DIMENSION;
	    o->frame_proportion = 0;
	}
	if(frame_dimension_isset){
	    if(frame_proportion_isset){
		fputs("Both dimension and proportion were given.  Use only one of --dimension=<num> or --proportion=<num>.\n", stderr);
		valid = false;
	    }else{
		/*  dimension ϡproportion  0 ˤơdimension ͤΤߤͭˤʤ褦  */
		o->frame_proportion = 0;
	    }
	}
	if(frame_proportion_isset){
	    /*  proportion ϡdimension  0 ˤơproportion ͤΤߤͭˤʤ褦  */
	    o->frame_dimension = 0;
	}
	if(!frame_sample_count_isset){
	    fprintf(stdout, "Sample count was not given.  Using default value %d.\n", DEFAULT_FRAME_SAMPLE_COUNT);
	    o->frame_sample_count = DEFAULT_FRAME_SAMPLE_COUNT;
	}
    }
    if(!valid){
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }

    NDD_RETURN;	
}

int main(int argc, char *argv[])
{
    NDD_START;

    /*  Initialize random seed  */
    NDD_IFOK{
	ndd_uint32 random_seed = 0;
	FILE *fp = fopen("/dev/random", "r");
	fread(&random_seed, sizeof(ndd_uint32), 1, fp);
	fclose(fp);
	srand(random_seed);
    }

    options opt = {NULL, NULL, DEFAULT_FRAME_PROPORTION, DEFAULT_FRAME_DIMENSION, DEFAULT_FRAME_SAMPLE_COUNT};
    NDD_VERIFY_RESULT(get_options(&opt, argc, argv));

    ndd_fvd *fvd = NULL;
    NDD_VERIFY_RESULT_OR(ndd_fvd_create(&fvd)){
	fputs("Failed to create fvd object.\n", stderr);
    }
    ndd_ffs *ffs = NULL;
    NDD_VERIFY_RESULT_OR(ndd_ffs_create(&ffs)){
	fputs("Failed to create ffs object.\n", stderr);
    }

//    FILE *input_fp = NULL;
//    NDD_VERIFY((input_fp = fopen(opt.input_fvdfile, "r")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
//    NDD_VERIFY_RESULT_OR(ndd_fvd_read_from_file(fvd, input_fp)){
    NDD_VERIFY_RESULT_OR(ndd_fvd_assign_file(fvd, opt.input_fvdfile, ndd_file_mode_read)){
	fputs("Failed to read fvd from file.\n", stderr);
    }
//    if(input_fp != NULL){
//	fclose(input_fp);
//    }
    
    NDD_IFOK{
    	ffs->header->frame_proportion   = opt.frame_proportion;
	ffs->header->frame_dimension    = opt.frame_dimension;
	ffs->header->frame_sample_count = opt.frame_sample_count;
    }
    NDD_VERIFY_RESULT_OR(ndd_create_ffs_from_fvd(ffs, fvd, ndd_callback_default_handler, NULL)){
	fputs("Failed to create ffs from fvd.\n", stderr);
    }
    
    FILE *output_fp = NULL;
    NDD_VERIFY((output_fp = fopen(opt.output_ffsfile, "w")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
    NDD_VERIFY_RESULT_OR(ndd_ffs_write_to_file(ffs, output_fp)){
	fputs("Failed to write ffs into file.\n", stderr);
    }

    if(ffs != NULL){
	ndd_ffs_destroy(&ffs);
    }
    if(fvd != NULL){
	ndd_fvd_destroy(&fvd);
    }
    if(output_fp != NULL){
	fclose(output_fp);
    }
    
    NDD_IFOK{
	return EXIT_SUCCESS;
    }else{
	fprintf(stderr, "Error code = %lu\n", __return_code);
	return EXIT_FAILURE;
    }
}

/*  end of file  */
