/*
    Copyright 2006-2009 Nagoya University

    This file is part of NDD (Near Duplicate Detector).

    NDD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    NDD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NDD.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ndd/fvd.h>
#include <ndd/error.h>
#include <ndd/mkfvd.h>
#include <ndd/file.h>
#include <ndd/defaults.h>

#include <getopt.h>
typedef struct __options
{
    const char *input_mpegfile;   /*  Ϥ MPEG ե̾  */
    const char *output_fvdfile;   /*  Ϥ FVD ե̾  */

    ndd_uint32 region_left;
    ndd_uint32 region_top;
    ndd_uint32 region_width;
    ndd_uint32 region_height;
    ndd_uint32 block_width;
    ndd_uint32 block_height;
    ndd_uint32 color_resolution;

} options;

static const struct option long_options[]={
    {"mpeg", required_argument, 0, 0},
    {"fvd", required_argument, 0, 0},
    {"region", required_argument, 0, 0},
    {"block", required_argument, 0, 0},
    {"color-resolution", required_argument, 0, 0},     

    {NULL, 0, 0, 0},
};

static ndd_result
show_usage()
{
    fputs("Usage: mkfvd  --mpeg=<input-mpegfile> --fvd=<output-fvdfile>\n", stderr);
    fputs("             [--region=<left>,<top>,<width>,<height>]\n", stderr);
    fputs("             [--block=<width>,<height>]\n", stderr);
    fputs("             [--color-resolution=<number>]\n", stderr);
    return NDD_SUCCESS;
}   

static ndd_result
get_options(options *o, int argc, char *argv[])
{
    NDD_START;

    ndd_bool input_mpegfile_isset      = false; 	
    ndd_bool output_fvdfile_isset      = false;
    ndd_bool region_isset	       = false;   
    ndd_bool block_isset	       = false;     
    ndd_bool color_resolution_isset    = false;  
   
    ndd_bool valid = true;
    NDD_WHILE(1){
	int index = -1;
	int c = getopt_long(argc, argv, "", long_options, &index);

	if(c == -1){
	    break;
	}
	switch(c){
	case '?':
	    show_usage();
	    valid = false;
	    break;

	default:
	    switch(index){	
	    case 0:
		o->input_mpegfile = optarg;
		input_mpegfile_isset = true;
		break;
	    case 1:
		o->output_fvdfile = optarg;
		output_fvdfile_isset = true;
		break;
	    
	    case 2:
	    	if(sscanf(optarg, "%lu,%lu,%lu,%lu", &o->region_left, &o->region_top, &o->region_width, &o->region_height) == 4){
		    region_isset = true;
	    	}else{
		    fputs("Invalid region value. Use --region=<left>,<top>,<width>,<height>.\n", stderr);
		    valid = false;
	    	}
	    	break;
	 
	    case 3:
	    	if(sscanf(optarg, "%lu,%lu", &o->block_width, &o->block_height) == 2){
		    block_isset = true;
	    	}else{
		    fputs("Invalid block value. Use --block=<width>,<height>.\n", stderr);
		    valid = false;
	    	}
	    	break;

	    case 4:
	   	o->color_resolution =atol(optarg) ;
	   	color_resolution_isset = false;  
	 	break;
	  
	    }
	}
    }

    /*  å  */
    if(valid){
	if(o->input_mpegfile == NULL){
	    fputs("No mepg filename was given.  Use --mpeg=<FILENAME>.\n", stderr);
	    NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
	}
	if(o->output_fvdfile == NULL){
	    fputs("No fvd filename was given.  Use --fvd=<FILENAME>.\n", stderr);
	    NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
	}
	if(o->region_width  % o->block_width  != 0 ||
	   o->region_height % o->block_height != 0){
	    fputs("Width/height of region must be multiple of width/height of block.\n", stderr);
	    NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
	}
    }
    if(!valid){
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }
    NDD_RETURN;
}

int main(int argc, char *argv[])
{
    NDD_START;

    options opt = {NULL, NULL,
		   0, 0, 0, 0,
		   DEFAULT_BLOCK_WIDTH, DEFAULT_BLOCK_HEIGHT,
		   DEFAULT_COLOR_RESOLUTION};
    NDD_VERIFY_RESULT(get_options(&opt, argc, argv));	

    ndd_fvd *fvd = NULL;
    NDD_VERIFY_RESULT_OR(ndd_fvd_create(&fvd)){
	fputs("Failed to create fvd object.\n", stderr);
    }

//    FILE *fp = NULL;
//    NDD_VERIFY((fp = fopen(opt.output_fvdfile, "w")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);

    NDD_VERIFY_RESULT(ndd_fvd_assign_file(fvd, opt.output_fvdfile, ndd_file_mode_write));

    NDD_IFOK{
    	fvd->header->region_left       = opt.region_left;
	fvd->header->region_top        = opt.region_top;
	fvd->header->region_width      = opt.region_width;
	fvd->header->region_height     = opt.region_height;
	fvd->header->block_width       = opt.block_width;
	fvd->header->block_height      = opt.block_height;
	fvd->header->color_resolution  = opt.color_resolution;
    }
    NDD_VERIFY_RESULT_OR(ndd_create_fvd_from_mpeg(fvd,opt.input_mpegfile, ndd_callback_default_handler, NULL)){
	fputs("Failed to create fvd from mpeg.\n", stderr);
    }

//    NDD_VERIFY_RESULT_OR(ndd_fvd_write_to_file(fvd, fp)){
//	fputs("Failed to write fvd into file.\n", stderr);
//    }

    if(fvd != NULL){
	ndd_fvd_destroy(&fvd);
    }
//    if(fp != NULL){
//	fclose(fp);
//    }
    
    NDD_IFOK{
	return EXIT_SUCCESS;
    }else{
	fprintf(stderr, "Error code = %lu\n", __return_code);
	return EXIT_FAILURE;
    }
}

/*  end of file  */
