/*
    Copyright 2006-2009 Nagoya University

    This file is part of NDD (Near Duplicate Detector).

    NDD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    NDD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NDD.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ndd/mksfd.h>
#include <ndd/sfd.h>
#include <ndd/sfs.h>
#include <ndd/ffd.h>
#include <ndd/error.h>
#include <ndd/file.h>
#include <ndd/defaults.h>

#include <getopt.h>
typedef struct __options
{
    const char *input_ffdfile;   /*  Ϥ FFD ե̾  */
    const char *input_sfsfile;   /*  Ϥ SFS ե̾  */
    const char *output_sfdfile;  /*  Ϥ SFD ե̾  */

} options;

static const struct option long_options[]={
    {"ffd", required_argument, 0, 0},
    {"sfs", required_argument, 0, 0},
    {"sfd", required_argument, 0, 0},
    {NULL, 0, 0, 0},
};

static ndd_result
show_usage()
{
    fputs("Usage: mksfd --ffd=<input-ffdfile> --sfs=<input-sfsfile> --sfd=<output-sfdfile>\n", stderr);
    return NDD_SUCCESS;
}   


static ndd_result
get_options(options *o, int argc, char *argv[])
{
    NDD_START;

    ndd_bool input_ffdfile_isset  = false;
    ndd_bool input_sfsfile_isset  = false;
    ndd_bool output_sfdfile_isset = false;
 
    ndd_bool valid = true;
    NDD_WHILE(1){
	int index = -1;
	int c = getopt_long(argc, argv, "", long_options, &index);

	if(c == -1){
	    break;
	}
	switch(c){
	case '?':
	    show_usage();
	    valid = false;
	    break;

	default:
	    switch(index){	
	    case 0:
		o->input_ffdfile = optarg;
		input_ffdfile_isset = true;
		break;
	    case 1:
		o->input_sfsfile = optarg;
		input_sfsfile_isset = true;
		break;
	    case 2:
		o->output_sfdfile = optarg;
		output_sfdfile_isset = true;
		break;
	    }
	}
    }

    /*  å  */
    if(valid){
	if(o->input_ffdfile == NULL){
	    fputs("No ffd filename was given.  Use --ffd=<FILENAME>.\n", stderr);
	    NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
	}
	if(o->input_sfsfile == NULL){
	    fputs("No sfs filename was given.  Use --sfs=<FILENAME>.\n", stderr);
	    NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
	}
	if(o->output_sfdfile == NULL){
	    fputs("No ffd filename was given.  Use --sfd=<FILENAME>.\n", stderr);
	    NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
	}
    }
    if(!valid){
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }
    
    NDD_RETURN;
}



int main(int argc, char *argv[])
{
    NDD_START;

    options opt = {NULL, NULL, NULL};
    NDD_VERIFY_RESULT(get_options(&opt, argc, argv));	


    ndd_ffd *ffd = NULL;
    NDD_VERIFY_RESULT_OR(ndd_ffd_create(&ffd)){
	fputs("Failed to create ffd object.\n", stderr);
    }
    ndd_sfs *sfs = NULL;
    NDD_VERIFY_RESULT_OR(ndd_sfs_create(&sfs)){
	fputs("Failed to create sfs object.\n", stderr);
    }
    ndd_sfd *sfd = NULL;
    NDD_VERIFY_RESULT_OR(ndd_sfd_create(&sfd)){
	fputs("Failed to create sfd object.\n", stderr);
    }

    FILE *input_fp = NULL;
    NDD_VERIFY((input_fp = fopen(opt.input_ffdfile, "r")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
    NDD_VERIFY_RESULT_OR(ndd_ffd_read_from_file(ffd, input_fp)){
	fputs("Failed to read ffd from file.\n", stderr);
    }
    if(input_fp != NULL){
	fclose(input_fp);
	input_fp = NULL;
    }

    NDD_VERIFY((input_fp = fopen(opt.input_sfsfile, "r")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
    NDD_VERIFY_RESULT_OR(ndd_sfs_read_from_file(sfs, input_fp)){
	fputs("Failed to read sfs from file.\n", stderr);
    }
    if(input_fp != NULL){
	fclose(input_fp);
	input_fp = NULL;
    }

    NDD_VERIFY_RESULT_OR(ndd_create_sfd_from_ffd(sfd, ffd, sfs, ndd_callback_default_handler, NULL)){
	fputs("Failed to create sfd from ffd.\n", stderr);
    }

    FILE *output_fp = NULL;
    NDD_VERIFY((output_fp = fopen(opt.output_sfdfile, "w")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
    NDD_VERIFY_RESULT_OR(ndd_sfd_write_to_file(sfd, output_fp)){
	fputs("Failed to write sfd into file.\n", stderr);
    }

    if(sfd != NULL){
	ndd_sfd_destroy(&sfd);
    }
    if(sfs != NULL){
	ndd_sfs_destroy(&sfs);
    }
    if(ffd != NULL){
	ndd_ffd_destroy(&ffd);
    }
    if(output_fp != NULL){
	fclose(output_fp);
    }
    
    NDD_IFOK{
	return EXIT_SUCCESS;
    }else{
	fprintf(stderr, "Error code = %lu\n", __return_code);
	return EXIT_FAILURE;
    }
}

/*  end of file  */
