/*
    Copyright 2006-2009 Nagoya University

    This file is part of NDD (Near Duplicate Detector).

    NDD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    NDD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NDD.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ndd/mksfs.h>
#include <ndd/sfs.h>
#include <ndd/ffd.h>
#include <ndd/error.h>
#include <ndd/file.h>
#include <ndd/defaults.h>

#include <getopt.h>
typedef struct __options
{
    const char *input_ffdfile;   /*  Ϥ FFD ե̾  */
    const char *output_sfsfile;  /*  Ϥ SFS ե̾  */

    ndd_uint32 segment_size;
    ndd_float8 segment_proportion;
    ndd_uint32 segment_dimension;
    ndd_uint32 segment_sample_count;

} options;

static const struct option long_options[]={
    {"ffd", required_argument, 0, 0},
    {"sfs", required_argument, 0, 0},
    {"size", required_argument, 0, 0},
    {"proportion", required_argument, 0, 0},
    {"dimension", required_argument, 0, 0},
    {"sample-count", required_argument, 0, 0},

    {NULL, 0, 0, 0},
};

static ndd_result
show_usage()
{
    fputs("Usage: mksfs  --ffd=<input-ffdfile> --sfs=<output-sfsfile>\n", stderr);
    fputs("             [--size=<number>]\n", stderr);
    fputs("             [--proportion=<number>]\n", stderr);
    fputs("             [--dimension=<number>]\n", stderr);
    fputs("             [--sample-count=<number>]\n", stderr);    
    return NDD_SUCCESS;
} 

static ndd_result
get_options(options *o, int argc, char *argv[])
{
    NDD_START;

    ndd_bool input_ffdfile_isset      = false;
    ndd_bool output_sfsfile_isset     = false;
    ndd_bool segment_size_isset    = false;
    ndd_bool segment_proportion_isset    = false;
    ndd_bool segment_dimension_isset = false;
    ndd_bool segment_sample_count_isset = false;

    ndd_bool valid = true;
    NDD_WHILE(1){
	int index = -1;
	int c = getopt_long(argc, argv, "", long_options, &index);

	if(c == -1){
	    break;
	}
	switch(c){
	case '?':
	    show_usage();
	    valid = false;
	    break;

	default:
	    switch(index){
	    case 0:
		o->input_ffdfile = optarg;
		input_ffdfile_isset = true;
		break;
	    case 1:
		o->output_sfsfile = optarg;
		output_sfsfile_isset = true;
		break;
	    case 2:
		o->segment_size = atol(optarg);
		segment_size_isset = true;
		break;
	    case 3:
		o->segment_proportion = atof(optarg);
		segment_proportion_isset = true;
		break;
            case 4:
		o->segment_dimension = atol(optarg);
		segment_dimension_isset = true;
		break;
            case 5:
		o->segment_sample_count = atol(optarg);
		segment_sample_count_isset = true;
		break;
	    }
	}
    }

    /*  å  */
    if(valid){
	if(!input_ffdfile_isset){
	    fputs("No ffd filename was given.  Use --ffd=<FILENAME>.\n", stderr);
	    valid = false;
	}
	if(!output_sfsfile_isset){
	    fputs("No sfs filename was given.  Use --sfs=<FILENAME>.\n", stderr);
	    valid = false;
	}
	if(!segment_dimension_isset && !segment_proportion_isset){
	    fprintf(stdout, "Dimension nor proportion was given.  Using default dimension size %d.\n", DEFAULT_SEGMENT_DIMENSION);
	    o->segment_dimension  = DEFAULT_SEGMENT_DIMENSION;
	    o->segment_proportion = 0;
	}
	if(segment_dimension_isset){
	    if(segment_proportion_isset){
		fputs("Both dimension and proportion were given.  Use only one of --dimension=<num> or --proportion=<num>.\n", stderr);
		valid = false;
	    }else{
		/*  dimension ϡproportion  0 ˤơdimension ͤΤߤͭˤʤ褦  */
		o->segment_proportion = 0;
	    }
	}
	if(segment_proportion_isset){
	    /*  proportion ϡdimension  0 ˤơproportion ͤΤߤͭˤʤ褦  */
	    o->segment_dimension = 0;
	}
	if(!segment_sample_count_isset){
	    fprintf(stdout, "Sample count was not given.  Using default value %d.\n", DEFAULT_SEGMENT_SAMPLE_COUNT);
	    o->segment_sample_count = DEFAULT_SEGMENT_SAMPLE_COUNT;
	}
    }
    if(!valid){
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }
    NDD_RETURN;	
}


int main(int argc, char *argv[])
{
    NDD_START;

    options opt = {NULL, NULL, DEFAULT_SEGMENT_SIZE, DEFAULT_SEGMENT_DIMENSION, DEFAULT_SEGMENT_PROPORTION, DEFAULT_SEGMENT_SAMPLE_COUNT};
    NDD_VERIFY_RESULT(get_options(&opt, argc, argv));

    
    ndd_ffd *ffd = NULL;
    NDD_VERIFY_RESULT_OR(ndd_ffd_create(&ffd)){
	fputs("Failed to create ffd object.\n", stderr);
    }
    ndd_sfs *sfs = NULL;
    NDD_VERIFY_RESULT_OR(ndd_sfs_create(&sfs)){
	fputs("Failed to create sfs object.\n", stderr);
    }

    FILE *input_fp = NULL;
    NDD_VERIFY((input_fp = fopen(opt.input_ffdfile, "r")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
    NDD_VERIFY_RESULT_OR(ndd_ffd_read_from_file(ffd, input_fp)){
	fputs("Failed to read ffd from file.\n", stderr);
    }
    if(input_fp != NULL){
	fclose(input_fp);
    }

     NDD_IFOK{
    	sfs->header->segment_size        = opt.segment_size;
	sfs->header->segment_proportion  = opt.segment_proportion;
	sfs->header->segment_dimension   = opt.segment_dimension;
	sfs->header->segment_sample_count= opt.segment_sample_count;
    }

    NDD_VERIFY_RESULT_OR(ndd_create_sfs_from_ffd(sfs, ffd, ndd_callback_default_handler, NULL)){
	fputs("Failed to create sfs from ffd.\n", stderr);
    }

    FILE *output_fp = NULL;
    NDD_VERIFY((output_fp = fopen(opt.output_sfsfile, "w")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
    NDD_VERIFY_RESULT_OR(ndd_sfs_write_to_file(sfs, output_fp)){
	fputs("Failed to write sfs into file.\n", stderr);
    }

    if(sfs != NULL){
	ndd_sfs_destroy(&sfs);
    }
    if(ffd != NULL){
	ndd_ffd_destroy(&ffd);
    }
    if(output_fp != NULL){
	fclose(output_fp);
    }
    
    NDD_IFOK{
	return EXIT_SUCCESS;
    }else{
	fprintf(stderr, "Error code = %lu\n", __return_code);
	return EXIT_FAILURE;
    }
}

/*  end of file  */
