/*
    Copyright 2006-2009 Nagoya University

    This file is part of NDD (Near Duplicate Detector).

    NDD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    NDD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NDD.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ndd/search.h>
#include <ndd/sfd.h>
#include <ndd/ffd.h>
#include <ndd/fvd.h>
#include <ndd/error.h>
#include <ndd/file.h>

#include <getopt.h>

typedef struct __options
{
    const char *input_sfdfile;
    const char *input_ffdfile;
    const char *input_fvdfile;

    ndd_float8 sfd_threshold;
    ndd_float8 ffd_threshold;
    ndd_float8 fvd_threshold;

    ndd_uint32 merge_level;
    ndd_uint32 minimum_segment_size;

} options;

static const struct option long_options[] = {
    {"sfd", required_argument, 0, 0},
    {"ffd", required_argument, 0, 0},
    {"fvd", required_argument, 0, 0},
    {"sfd-threshold", required_argument, 0, 0},
    {"ffd-threshold", required_argument, 0, 0},
    {"fvd-threshold", required_argument, 0, 0},
    {"merge-level", required_argument, 0, 0},
    {"minimum-segment-size", required_argument, 0, 0},
    {NULL, 0, 0, 0},
};

static ndd_result
show_usage()
{
    fputs("Usage: search --sfd=<input-sfdfile> [--sfd-threshold=<distance>]\n", stderr);
    fputs("              --ffd=<input-ffdfile> [--ffd-threshold=<distance>]\n", stderr);
    fputs("              --fvd=<input-fvdfile> [--fvd-threshold=<distance>]\n", stderr);
    fputs("             [--merge-level={1|2|3}]\n", stderr);
    fputs("             [--minimum-segment-size=<frame-count>]\n", stderr);
    return NDD_SUCCESS;
}

static ndd_result
get_options(options *o, int argc, char *argv[])
{
    NDD_START;

    NDD_WHILE(1){
	int index = -1;
	int c = getopt_long(argc, argv, "", long_options, &index);

	if(c == -1){
	    break;
	}
	switch(c){
	case '?':
	    show_usage();
	    NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
	    break;

	default:
	    switch(index){
	    case 0:
		o->input_sfdfile = optarg;
		break;
	    case 1:
		o->input_ffdfile = optarg;
		break;
	    case 2:
		o->input_fvdfile = optarg;
		break;
	    case 3:
		o->sfd_threshold = atof(optarg);
		break;
	    case 4:
		o->ffd_threshold = atof(optarg);
		break;
	    case 5:
		o->fvd_threshold = atof(optarg);
		break;
	    case 6:
		{
		    ndd_uint32 level = atol(optarg);
		    switch(level){
		    case 1:  o->merge_level = ndd_search_merge_level_none;  break;
		    case 2:  o->merge_level = ndd_search_merge_level_line;  break;
		    case 3:  o->merge_level = ndd_search_merge_level_square;  break;
		    default:
			fputs("Invalid merge level.  Use --merge-level=[1|2|3].\n", stderr);
			NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
		    }
		}
		break;
	    case 7:
		{
		    o->minimum_segment_size = atol(optarg);
		}
		break;
	    }
	}
    }

    /*  å  */
    if(o->input_sfdfile == NULL){
	fputs("No sfd filename was given.  Use --sfd=<FILENAME>.\n", stderr);
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }else
    if(o->input_ffdfile == NULL){
	fputs("No ffd filename was given.  Use --ffd=<FILENAME>.\n", stderr);
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }else
    if(o->input_fvdfile == NULL){
	fputs("No fvd filename was given.  Use --fvd=<FILENAME>.\n", stderr);
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }else{
    }

    NDD_RETURN;
}

int main(int argc, char *argv[])
{
    NDD_START;

    options opt = {NULL, NULL, NULL, 0.8, 0.8, 0.8, ndd_search_merge_level_default};
    NDD_VERIFY_RESULT(get_options(&opt, argc, argv));

    ndd_fvd *fvd = NULL;
    NDD_VERIFY_RESULT_OR(ndd_fvd_create(&fvd)){
	fputs("Failed to create fvd object.\n", stderr);
    }
    ndd_ffd *ffd = NULL;
    NDD_VERIFY_RESULT_OR(ndd_ffd_create(&ffd)){
	fputs("Failed to create ffd object.\n", stderr);
    }
    ndd_sfd *sfd = NULL;
    NDD_VERIFY_RESULT_OR(ndd_sfd_create(&sfd)){
	fputs("Failed to create sfd object.\n", stderr);
    }

    FILE *input_fp = NULL;
    NDD_VERIFY((input_fp = fopen(opt.input_sfdfile, "r")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
    NDD_VERIFY_RESULT_OR(ndd_sfd_read_from_file(sfd, input_fp)){
	fputs("Failed to read sfd from file.\n", stderr);
    }
    if(input_fp != NULL){
	fclose(input_fp);
	input_fp = NULL;
    }

    NDD_VERIFY((input_fp = fopen(opt.input_ffdfile, "r")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
    NDD_VERIFY_RESULT_OR(ndd_ffd_read_from_file(ffd, input_fp)){
	fputs("Failed to read ffd from file.\n", stderr);
    }
    if(input_fp != NULL){
	fclose(input_fp);
	input_fp = NULL;
    }

//    NDD_VERIFY((input_fp = fopen(opt.input_fvdfile, "r")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
//    NDD_VERIFY_RESULT_OR(ndd_fvd_read_from_file(fvd, input_fp)){
    NDD_VERIFY_RESULT_OR(ndd_fvd_assign_file(fvd, opt.input_fvdfile, ndd_file_mode_read)){
	fputs("Failed to read fvd from file.\n", stderr);
    }
//    if(input_fp != NULL){
//	fclose(input_fp);
//	input_fp = NULL;
//    }

    NDD_IFOK{
	sfd->header->sfd_search_threshold = opt.sfd_threshold;
	ffd->header->ffd_search_threshold = opt.ffd_threshold;
	fvd->header->fvd_search_threshold = opt.fvd_threshold;
	sfd->header->minimum_segment_size = opt.minimum_segment_size;
    }
    NDD_VERIFY_RESULT(ndd_search(sfd, ffd, fvd, sfd, ffd, fvd, opt.merge_level,
				 ndd_callback_default_handler, NULL));

    if(sfd != NULL){
	ndd_sfd_destroy(&sfd);
    }
    if(ffd != NULL){
	ndd_ffd_destroy(&ffd);
    }
    if(fvd != NULL){
	ndd_fvd_destroy(&fvd);
    }
    
    NDD_IFOK{
	return EXIT_SUCCESS;
    }else{
	fprintf(stderr, "Error code = %lu\n", __return_code);
	return EXIT_FAILURE;
    }
}

/*  end of file  */
