/*
    Copyright 2006-2009 Nagoya University

    This file is part of NDD (Near Duplicate Detector).

    NDD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    NDD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NDD.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ndd/search.h>
#include <ndd/sfd.h>
#include <ndd/ffd.h>
#include <ndd/fvd.h>
#include <ndd/error.h>
#include <ndd/file.h>

#include <getopt.h>

typedef struct __options
{
    const char *input_sfdfile1;
    const char *input_ffdfile1;
    const char *input_fvdfile1;
    const char *input_sfdfile2;
    const char *input_ffdfile2;
    const char *input_fvdfile2;

    ndd_float8 sfd_threshold;
    ndd_float8 ffd_threshold;
    ndd_float8 fvd_threshold;

    ndd_uint32 merge_level;
    ndd_uint32 minimum_segment_size;

} options;

static const struct option long_options[] = {
    {"sfd1", required_argument, 0, 0},
    {"ffd1", required_argument, 0, 0},
    {"fvd1", required_argument, 0, 0},
    {"sfd2", required_argument, 0, 0},
    {"ffd2", required_argument, 0, 0},
    {"fvd2", required_argument, 0, 0},
    {"sfd-threshold", required_argument, 0, 0},
    {"ffd-threshold", required_argument, 0, 0},
    {"fvd-threshold", required_argument, 0, 0},
    {"merge-level", required_argument, 0, 0},
    {"minimum-segment-size", required_argument, 0, 0},
    {NULL, 0, 0, 0},
};

static ndd_result
show_usage()
{
    fputs("Usage: search --sfd1=<input-sfdfile1> --sfd2=<input-sfdfile2>\n", stderr);
    fputs("              --ffd1=<input-ffdfile1> --ffd2=<input-ffdfile2>\n", stderr);
    fputs("              --fvd1=<input-fvdfile1> --fvd2=<input-fvdfile2>\n", stderr);
    fputs("             [--sfd-threshold=<distance>]\n", stderr);
    fputs("             [--ffd-threshold=<distance>]\n", stderr);
    fputs("             [--fvd-threshold=<distance>]\n", stderr);
    fputs("             [--merge-level={1|2|3}]\n", stderr);
    fputs("             [--minimum-segment-size=<frame-count>]\n", stderr);
    return NDD_SUCCESS;
}

static ndd_result
get_options(options *o, int argc, char *argv[])
{
    NDD_START;

    NDD_WHILE(1){
	int index = -1;
	int c = getopt_long(argc, argv, "", long_options, &index);

	if(c == -1){
	    break;
	}
	switch(c){
	case '?':
	    show_usage();
	    NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
	    break;

	default:
	    switch(index){
	    case 0:
		o->input_sfdfile1 = optarg;
		break;
	    case 1:
		o->input_ffdfile1 = optarg;
		break;
	    case 2:
		o->input_fvdfile1 = optarg;
		break;
	    case 3:
		o->input_sfdfile2 = optarg;
		break;
	    case 4:
		o->input_ffdfile2 = optarg;
		break;
	    case 5:
		o->input_fvdfile2 = optarg;
		break;
	    case 6:
		o->sfd_threshold = atof(optarg);
		break;
	    case 7:
		o->ffd_threshold = atof(optarg);
		break;
	    case 8:
		o->fvd_threshold = atof(optarg);
		break;
	    case 9:
		{
		    ndd_uint32 level = atol(optarg);
		    switch(level){
		    case 1:  o->merge_level = ndd_search_merge_level_none;  break;
		    case 2:  o->merge_level = ndd_search_merge_level_line;  break;
		    case 3:  o->merge_level = ndd_search_merge_level_square;  break;
		    default:
			fputs("Invalid merge level.  Use --merge-level=[1|2|3].\n", stderr);
			NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
		    }
		}
		break;
	    case 10:
		{
		    o->minimum_segment_size = atol(optarg);
		}
		break;
	    }
	}
    }

    /*  å  */
    if(o->input_sfdfile1 == NULL){
	fputs("No sfd filename was given.  Use --sfd1=<FILENAME>.\n", stderr);
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }else
    if(o->input_ffdfile1 == NULL){
	fputs("No ffd filename was given.  Use --ffd1=<FILENAME>.\n", stderr);
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }else
    if(o->input_fvdfile1 == NULL){
	fputs("No fvd filename was given.  Use --fvd1=<FILENAME>.\n", stderr);
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }else
    if(o->input_sfdfile2 == NULL){
	fputs("No sfd filename was given.  Use --sfd2=<FILENAME>.\n", stderr);
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }else
    if(o->input_ffdfile2 == NULL){
	fputs("No ffd filename was given.  Use --ffd2=<FILENAME>.\n", stderr);
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }else
    if(o->input_fvdfile2 == NULL){
	fputs("No fvd filename was given.  Use --fvd2=<FILENAME>.\n", stderr);
	NDD_SET_ERROR(NDD_ERROR_APPLICATION_ERROR);
    }else{
    }

    NDD_RETURN;
}

int main(int argc, char *argv[])
{
    NDD_START;

    options opt = {NULL, NULL, NULL, NULL, NULL, NULL, 0.8, 0.8, 0.8, ndd_search_merge_level_default};
    NDD_VERIFY_RESULT(get_options(&opt, argc, argv));

    ndd_fvd *fvd1 = NULL;
    NDD_VERIFY_RESULT_OR(ndd_fvd_create(&fvd1)){
	fputs("Failed to create fvd object.\n", stderr);
    }
    ndd_ffd *ffd1 = NULL;
    NDD_VERIFY_RESULT_OR(ndd_ffd_create(&ffd1)){
	fputs("Failed to create ffd object.\n", stderr);
    }
    ndd_sfd *sfd1 = NULL;
    NDD_VERIFY_RESULT_OR(ndd_sfd_create(&sfd1)){
	fputs("Failed to create sfd object.\n", stderr);
    }
    ndd_fvd *fvd2 = NULL;
    NDD_VERIFY_RESULT_OR(ndd_fvd_create(&fvd2)){
	fputs("Failed to create fvd object.\n", stderr);
    }
    ndd_ffd *ffd2 = NULL;
    NDD_VERIFY_RESULT_OR(ndd_ffd_create(&ffd2)){
	fputs("Failed to create ffd object.\n", stderr);
    }
    ndd_sfd *sfd2 = NULL;
    NDD_VERIFY_RESULT_OR(ndd_sfd_create(&sfd2)){
	fputs("Failed to create sfd object.\n", stderr);
    }

    FILE *input_fp = NULL;
    NDD_VERIFY((input_fp = fopen(opt.input_sfdfile1, "r")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
    NDD_VERIFY_RESULT_OR(ndd_sfd_read_from_file(sfd1, input_fp)){
	fputs("Failed to read sfd from file.\n", stderr);
    }
    if(input_fp != NULL){
	fclose(input_fp);
	input_fp = NULL;
    }

    NDD_VERIFY((input_fp = fopen(opt.input_ffdfile1, "r")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
    NDD_VERIFY_RESULT_OR(ndd_ffd_read_from_file(ffd1, input_fp)){
	fputs("Failed to read ffd from file.\n", stderr);
    }
    if(input_fp != NULL){
	fclose(input_fp);
	input_fp = NULL;
    }

//    NDD_VERIFY((input_fp = fopen(opt.input_fvdfile1, "r")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
//    NDD_VERIFY_RESULT_OR(ndd_fvd_read_from_file(fvd1, input_fp)){
    NDD_VERIFY_RESULT_OR(ndd_fvd_assign_file(fvd1, opt.input_fvdfile1, ndd_file_mode_read)){
	fputs("Failed to read fvd from file.\n", stderr);
    }
//    if(input_fp != NULL){
//	fclose(input_fp);
//	input_fp = NULL;
//    }
    NDD_VERIFY((input_fp = fopen(opt.input_sfdfile2, "r")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
    NDD_VERIFY_RESULT_OR(ndd_sfd_read_from_file(sfd2, input_fp)){
	fputs("Failed to read sfd from file.\n", stderr);
    }
    if(input_fp != NULL){
	fclose(input_fp);
	input_fp = NULL;
    }

    NDD_VERIFY((input_fp = fopen(opt.input_ffdfile2, "r")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
    NDD_VERIFY_RESULT_OR(ndd_ffd_read_from_file(ffd2, input_fp)){
	fputs("Failed to read ffd from file.\n", stderr);
    }
    if(input_fp != NULL){
	fclose(input_fp);
	input_fp = NULL;
    }

//    NDD_VERIFY((input_fp = fopen(opt.input_fvdfile2, "r")) != NULL, NDD_ERROR_FILE_OPEN_FAILED);
//    NDD_VERIFY_RESULT_OR(ndd_fvd_read_from_file(fvd2, input_fp)){
    NDD_VERIFY_RESULT_OR(ndd_fvd_assign_file(fvd2, opt.input_fvdfile2, ndd_file_mode_read)){
	fputs("Failed to read fvd from file.\n", stderr);
    }
//    if(input_fp != NULL){
//	fclose(input_fp);
//	input_fp = NULL;
//    }

    NDD_IFOK{
	sfd1->header->sfd_search_threshold = opt.sfd_threshold;
	ffd1->header->ffd_search_threshold = opt.ffd_threshold;
	fvd1->header->fvd_search_threshold = opt.fvd_threshold;
	sfd1->header->minimum_segment_size = opt.minimum_segment_size;
    }
    NDD_VERIFY_RESULT(ndd_search(sfd1, ffd1, fvd1, sfd2, ffd2, fvd2,
				 opt.merge_level, ndd_callback_default_handler, NULL));

    if(sfd1 != NULL){
	ndd_sfd_destroy(&sfd1);
    }
    if(ffd1 != NULL){
	ndd_ffd_destroy(&ffd1);
    }
    if(fvd1 != NULL){
	ndd_fvd_destroy(&fvd1);
    }
    if(sfd2 != NULL){
	ndd_sfd_destroy(&sfd2);
    }
    if(ffd2 != NULL){
	ndd_ffd_destroy(&ffd2);
    }
    if(fvd2 != NULL){
	ndd_fvd_destroy(&fvd2);
    }
    
    NDD_IFOK{
	return EXIT_SUCCESS;
    }else{
	fprintf(stderr, "Error code = %lu\n", __return_code);
	return EXIT_FAILURE;
    }
}

/*  end of file  */
