/*
    Copyright 2006-2009 Nagoya University

    This file is part of NDD (Near Duplicate Detector).

    NDD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    NDD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NDD.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ndd/ffd.h>
#include <ndd/error.h>
#include <ndd/memory.h>
#include <ndd/file.h>
#include <ndd/ffs.h>

#include <string.h>

ndd_result
ndd_ffd_create(ndd_ffd **ffd)
{ 
    NDD_START;
    NDD_VERIFY_POINTER(ffd);

    /*  ΰ  */
    ndd_ffd *new_ffd = NULL;
    NDD_VERIFY_RESULT(ndd_calloc(&new_ffd, sizeof(ndd_ffd)));

    /*  ѿν  */
    NDD_VERIFY_RESULT(ndd_file_header_create(&new_ffd->header));
    NDD_VERIFY_RESULT(ndd_ffs_data_create   (&new_ffd->ffs));
    NDD_VERIFY_RESULT(ndd_ffd_data_create   (&new_ffd->ffd));

    /*  ̤Ǽ  */
    NDD_IFOK{
	*ffd = new_ffd;
	new_ffd = NULL;
    }

    /*  פʥǡդ  */
    if(new_ffd != NULL){
	ndd_ffd_destroy(&new_ffd);
    }

    NDD_RETURN;
}

ndd_result
ndd_ffd_destroy(ndd_ffd **ffd)
{
    NDD_START;
    NDD_VERIFY_POINTER(ffd);

    ndd_ffd *old_ffd;
    NDD_IFOK{
	old_ffd = *ffd;
    }

    NDD_VERIFY_RESULT(ndd_ffd_data_destroy   (&old_ffd->ffd, old_ffd->header));
    NDD_VERIFY_RESULT(ndd_ffs_data_destroy   (&old_ffd->ffs, old_ffd->header));
    NDD_VERIFY_RESULT(ndd_file_header_destroy(&old_ffd->header));
    NDD_VERIFY_RESULT(ndd_free(&old_ffd));
    
    NDD_IFOK{
	*ffd = NULL;
    }

    NDD_RETURN;
}

ndd_result
ndd_ffd_copy(ndd_ffd *dst_ffd, const ndd_ffd *src_ffd)
{
    NDD_START;
    NDD_VERIFY_POINTER(dst_ffd);
    NDD_VERIFY_POINTER(src_ffd);

    /*  إå򥳥ԡ  */
    ndd_file_header *new_header = NULL;
    NDD_VERIFY_RESULT(ndd_file_header_create(&new_header));
    NDD_VERIFY_RESULT(ndd_file_header_copy(new_header, src_ffd->header));

    /*  FFS ǡ򥳥ԡ  */
    ndd_ffs_data *new_ffs_data = NULL;
    NDD_VERIFY_RESULT(ndd_ffs_data_create(&new_ffs_data));
    NDD_VERIFY_RESULT(ndd_ffs_data_copy(new_ffs_data, src_ffd->ffs, src_ffd->header));

    /*  FFD ǡ򥳥ԡ  */
    ndd_ffd_data *new_ffd_data = NULL;
    NDD_VERIFY_RESULT(ndd_ffd_data_create(&new_ffd_data));
    NDD_VERIFY_RESULT(ndd_ffd_data_copy(new_ffd_data, src_ffd->ffd, src_ffd->header));

    /*  ǡǼ  */
    NDD_IFOK{
	if(dst_ffd->ffd != NULL){
	    NDD_VERIFY_RESULT(ndd_ffd_data_destroy(&dst_ffd->ffd, dst_ffd->header));
	}
	if(dst_ffd->ffs != NULL){
	    NDD_VERIFY_RESULT(ndd_ffs_data_destroy(&dst_ffd->ffs, dst_ffd->header));
	}
	if(dst_ffd->header != NULL){
	    NDD_VERIFY_RESULT(ndd_file_header_destroy(&dst_ffd->header));
	}
    }
    NDD_IFOK{
	dst_ffd->header = new_header;
	dst_ffd->ffd    = new_ffd_data;
	dst_ffd->ffs    = new_ffs_data;
	new_header   = NULL;
	new_ffd_data = NULL;
	new_ffs_data = NULL;
    }

    if(new_ffd_data != NULL){  ndd_ffd_data_destroy   (&new_ffd_data, new_header);  }
    if(new_ffs_data != NULL){  ndd_ffs_data_destroy   (&new_ffs_data, new_header);  }
    if(new_header   != NULL){  ndd_file_header_destroy(&new_header);    }

    NDD_RETURN;
}

ndd_result
ndd_ffd_cut(ndd_ffd *dst_ffd, const ndd_ffd *src_ffd, ndd_uint32 frame_offset, ndd_uint32 frame_count)
{
    NDD_START;
    NDD_VERIFY_POINTER(dst_ffd);
    NDD_VERIFY_POINTER(src_ffd);

    NDD_IFOK{
	if(dst_ffd == src_ffd){
	    NDD_SET_ERROR(NDD_ERROR_SAME_DATA);
	}else if(dst_ffd->header->frame_count > 0){
	    NDD_SET_ERROR(NDD_ERROR_DATA_ALREADY_EXISTS);
	}else if(frame_offset < src_ffd->header->frame_offset){
	    NDD_SET_ERROR(NDD_ERROR_FRAME_OFFSET_OUT_OF_BOUND);
	}else if(frame_offset + frame_count > src_ffd->header->frame_offset + src_ffd->header->frame_count){
	    NDD_SET_ERROR(NDD_ERROR_FRAME_COUNT_OUT_OF_BOUND);
	}
    }

    ndd_uint32 elements_per_frame = 0;
    NDD_VERIFY_RESULT(ndd_file_header_get_elements_per_frame(src_ffd->header, &elements_per_frame));
    
    ndd_float8 **new_frame_vector = NULL;
    NDD_VERIFY_RESULT(ndd_calloc(&new_frame_vector, sizeof(ndd_float8 *) * frame_count));

    ndd_uint32 i;
    NDD_FOR(i = 0; i < frame_count; i++){
	ndd_float8 *f = NULL;
	NDD_VERIFY_RESULT(ndd_malloc(&f, sizeof(ndd_float8) * elements_per_frame));
	NDD_IFOK{
	    memcpy(f, src_ffd->ffd->frame_vector[frame_offset - src_ffd->header->frame_offset + i],
		   sizeof(ndd_float8) * elements_per_frame);
	}

	NDD_IFOK{
	    new_frame_vector[i] = f;
	    f = NULL;
	}

	if(f != NULL){ ndd_free(&f); }
    }

    NDD_VERIFY_RESULT(ndd_file_header_copy(dst_ffd->header, src_ffd->header));
    NDD_VERIFY_RESULT(ndd_ffs_data_copy   (dst_ffd->ffs,    src_ffd->ffs, src_ffd->header));
    NDD_IFOK{
	dst_ffd->header->frame_offset = frame_offset;
	dst_ffd->header->frame_count  = frame_count;
	dst_ffd->ffd->frame_vector    = new_frame_vector;
	new_frame_vector = NULL;
    }

    if(new_frame_vector != NULL){
	for(i = 0; i < frame_count; i++){
	    if(new_frame_vector[i] != NULL){
		ndd_free(&new_frame_vector[i]);
	    }
	}
	ndd_free(&new_frame_vector);
    }

    NDD_RETURN;
}

ndd_result
ndd_ffd_concatenate(ndd_ffd *dst_ffd, const ndd_ffd *src_ffd_1, const ndd_ffd *src_ffd_2)
{
    NDD_START;
    NDD_VERIFY_POINTER(dst_ffd);
    NDD_VERIFY_POINTER(src_ffd_1);
    NDD_VERIFY_POINTER(src_ffd_2);

    NDD_IFOK{
	if(dst_ffd == src_ffd_2){
	    NDD_SET_ERROR(NDD_ERROR_SAME_DATA);
	}else if(dst_ffd->header->frame_count > 0 && dst_ffd != src_ffd_1){
	    NDD_SET_ERROR(NDD_ERROR_DATA_ALREADY_EXISTS);
	}else if(dst_ffd->header->frame_count > 0 &&
		 src_ffd_1->header->frame_offset + src_ffd_1->header->frame_count != src_ffd_2->header->frame_offset){
	    NDD_SET_ERROR(NDD_ERROR_FRAME_DATA_NOT_CONTIGUOUS);
	}
    }

    NDD_IFOK{
	if(dst_ffd != src_ffd_1){
	    NDD_VERIFY_RESULT(ndd_ffd_copy(dst_ffd, src_ffd_1));
	}
	if(dst_ffd == src_ffd_1 && dst_ffd->header->frame_count == 0){
	    return ndd_ffd_copy(dst_ffd, src_ffd_2);
	}
    }

    ndd_uint32 elements_per_frame = 0;
    NDD_VERIFY_RESULT(ndd_file_header_get_elements_per_frame(src_ffd_1->header, &elements_per_frame));
    
    ndd_uint32 i;
    NDD_FOR(i = 0; i < src_ffd_2->header->frame_count; i++){
	ndd_float8 *f = NULL;
	NDD_VERIFY_RESULT(ndd_malloc(&f, sizeof(ndd_float8) * elements_per_frame));
	NDD_IFOK{
	    memcpy(f, src_ffd_2->ffd->frame_vector[i], sizeof(ndd_float8) * elements_per_frame);
	}

	NDD_VERIFY_RESULT(ndd_realloc(&dst_ffd->ffd->frame_vector,
				      sizeof(ndd_float8 *) * (dst_ffd->header->frame_count + 1)));
	NDD_IFOK{
	    dst_ffd->ffd->frame_vector[dst_ffd->header->frame_count++] = f;
	    f = NULL;
	}

	if(f != NULL){ ndd_free(&f); }
    }

    NDD_RETURN;
}

ndd_result
ndd_ffd_read_from_file(ndd_ffd *ffd, FILE *fp)
{
    NDD_START;
    NDD_VERIFY_POINTER(ffd);
    NDD_VERIFY_POINTER(fp);

    /*  إåɤ߹  */
    ndd_file_header *new_header = NULL;
    NDD_VERIFY_RESULT(ndd_file_header_create(&new_header));
    NDD_VERIFY_RESULT(ndd_file_header_read_from_file(new_header, fp));

    /*  FFS ǡɤ߹  */
    ndd_ffs_data *new_ffs_data = NULL;
    NDD_VERIFY_RESULT(ndd_ffs_data_create(&new_ffs_data));
    NDD_VERIFY_RESULT(ndd_ffs_data_read_from_file(new_ffs_data, new_header, fp));

    /*  FFD ǡɤ߹  */
    ndd_ffd_data *new_ffd_data = NULL;
    NDD_VERIFY_RESULT(ndd_ffd_data_create(&new_ffd_data));
    NDD_VERIFY_RESULT(ndd_ffd_data_read_from_file(new_ffd_data, new_header, fp));

    /*  ǡǼ  */
    NDD_IFOK{
	if(ffd->ffs != NULL){
	    NDD_VERIFY_RESULT(ndd_ffs_data_destroy(&ffd->ffs, ffd->header));
	}
	if(ffd->ffd != NULL){
	    NDD_VERIFY_RESULT(ndd_ffd_data_destroy(&ffd->ffd, ffd->header));
	}
	if(ffd->header != NULL){
	    NDD_VERIFY_RESULT(ndd_file_header_destroy(&ffd->header));
	}
    }
    NDD_IFOK{
	ffd->header = new_header;
	ffd->ffs    = new_ffs_data;
	ffd->ffd    = new_ffd_data;
	new_header   = NULL;
	new_ffs_data = NULL;
	new_ffd_data = NULL;
    }

    if(new_ffs_data != NULL){  ndd_ffs_data_destroy   (&new_ffs_data, new_header);  }
    if(new_ffd_data != NULL){  ndd_ffd_data_destroy   (&new_ffd_data, new_header);  }
    if(new_header   != NULL){  ndd_file_header_destroy(&new_header);    }

    NDD_RETURN;
}

ndd_result
ndd_ffd_write_to_file(const ndd_ffd *ffd, FILE *fp)
{
    NDD_START;
    NDD_VERIFY_POINTER(ffd);
    NDD_VERIFY_POINTER(fp);

    /*  եإå񤭹  */
    NDD_VERIFY_RESULT(ndd_file_header_write_to_file(ffd->header, fp));

    /*  FFS ǡ񤭹  */
    NDD_VERIFY_RESULT(ndd_ffs_data_write_to_file(ffd->ffs, ffd->header, fp));

    /*  FFD ǡ񤭹  */
    NDD_VERIFY_RESULT(ndd_ffd_data_write_to_file(ffd->ffd, ffd->header, fp));

    NDD_RETURN;
}

ndd_result
ndd_ffd_print_to_file(const ndd_ffd *ffd, FILE *fp)
{
    NDD_START;
    NDD_VERIFY_POINTER(ffd);
    NDD_VERIFY_POINTER(fp);

    /*  եإå񤭹  */
    NDD_VERIFY_RESULT(ndd_file_header_print_to_file(ffd->header, fp));

    /*  FFS ǡ񤭹  */
    NDD_VERIFY_RESULT(ndd_ffs_data_print_to_file(ffd->ffs, ffd->header, fp));

    /*  FFD ǡ񤭹  */
    NDD_VERIFY_RESULT(ndd_ffd_data_print_to_file(ffd->ffd, ffd->header, fp));

    NDD_RETURN;
}

ndd_result
ndd_ffd_compare_format(const ndd_ffd *ffd1, const ndd_ffd *ffd2, ndd_bool *is_same)
{
    NDD_START;
    NDD_VERIFY_POINTER(ffd1);
    NDD_VERIFY_POINTER(ffd2);
    NDD_VERIFY_POINTER(is_same);
    
    NDD_VERIFY_RESULT(ndd_ffd_header_compare_format(ffd1->header, ffd2->header, is_same));
    NDD_VERIFY_RESULT(ndd_ffs_data_compare(ffd1->ffs, ffd1->header,
					   ffd2->ffs, ffd2->header, is_same));
    
    NDD_RETURN;
}

ndd_result
ndd_ffd_header_compare_format(const ndd_file_header *header1, const ndd_file_header *header2, ndd_bool *is_same)
{
    NDD_START;
    NDD_VERIFY_POINTER(header1);
    NDD_VERIFY_POINTER(header2);
    NDD_VERIFY_POINTER(is_same);

    ndd_bool same = false;
    NDD_VERIFY_RESULT(ndd_ffs_header_compare_format(header1, header2, &same));

    NDD_IFOK{
	*is_same = same;
    }

    NDD_RETURN;
}

ndd_result
ndd_ffd_data_create(ndd_ffd_data **ffd_data)
{ 
    NDD_START;
    NDD_VERIFY_POINTER(ffd_data);

    /*  ΰ  */
    ndd_ffd_data *new_ffd_data = NULL;
    NDD_VERIFY_RESULT(ndd_calloc(&new_ffd_data, sizeof(ndd_ffd_data)));

    /*  ѿν  */

    /*  ̤Ǽ  */
    NDD_IFOK{
	*ffd_data = new_ffd_data;
	new_ffd_data = NULL;
    }

    /*  פʥǡդ  */
    if(new_ffd_data != NULL){
	ndd_free(&new_ffd_data);
    }

    NDD_RETURN;
}

ndd_result
ndd_ffd_data_destroy(ndd_ffd_data **ffd_data, const ndd_file_header *header)
{
    NDD_START;
    NDD_VERIFY_POINTER(ffd_data);
    NDD_VERIFY_POINTER(header);

    ndd_ffd_data *old_ffd_data;
    NDD_IFOK{
	old_ffd_data = *ffd_data;
    }

    ndd_uint32 frame_count = 0;
    NDD_VERIFY_RESULT(ndd_file_header_get_frame_count(header, &frame_count));

    ndd_uint32 i;
    NDD_FOR(i = 0; i < frame_count; i++){
	NDD_VERIFY_RESULT(ndd_free(&old_ffd_data->frame_vector[i]));
    }
    NDD_VERIFY_RESULT(ndd_free(&old_ffd_data->frame_vector));
    NDD_VERIFY_RESULT(ndd_free(&old_ffd_data));
    
    NDD_IFOK{
	*ffd_data = NULL;
    }

    NDD_RETURN;
}

ndd_result
ndd_ffd_data_read_from_file(ndd_ffd_data *ffd_data, const ndd_file_header *header, FILE *fp)
{
    NDD_START;

    /*  ե졼礭ʤɤ  */
    ndd_uint32 frame_count =0;
    NDD_VERIFY_RESULT(ndd_file_header_get_frame_count(header, &frame_count));
 
     /*  ɤ߹ ffd ǡǼΰ  */
    ndd_float8 **new_vector_data = NULL;
    NDD_VERIFY_RESULT(ndd_calloc(&new_vector_data, sizeof(ndd_float8 *) * frame_count));

    /*  ƥե졼Υǡɤ߹  */
    ndd_uint32 i;
    NDD_FOR(i = 0 ; i < frame_count; i++){

	ndd_float8 *vector_data = NULL;
	NDD_VERIFY_RESULT(ndd_malloc(&vector_data, sizeof(ndd_float8) * header->frame_dimension));

	NDD_VERIFY(fread(vector_data, sizeof(ndd_float8), header->frame_dimension, fp) == header->frame_dimension,
		   NDD_ERROR_FILE_READ_FAILED);
	
	NDD_IFOK{
	    new_vector_data[i] = vector_data;
	    vector_data = NULL;
	}

	if(vector_data != NULL){
	    ndd_free(&vector_data);
	}	    
    }

    /*  줿ǡѿ˳Ǽ  */
    NDD_IFOK{
	ffd_data->frame_vector = new_vector_data;
	new_vector_data = NULL;
    }

    /*  Իʤɤθդ  */
    if(new_vector_data != NULL){
	for(i = 0 ; i < frame_count; i++){
	    ndd_free(&new_vector_data[i]);
	}
	ndd_free(&new_vector_data);
    }

    NDD_RETURN;
}

ndd_result
ndd_ffd_data_write_to_file(const ndd_ffd_data *ffd_data, const ndd_file_header *header, FILE *fp)
{
    NDD_START;

    ndd_uint32 frame_count = 0;
    NDD_VERIFY_RESULT(ndd_file_header_get_frame_count(header, &frame_count));
 	
    /*  ե졼ǡ񤭹  */
    ndd_uint32 i;
    NDD_FOR(i = 0 ; i < frame_count; i++){
	NDD_VERIFY(fwrite(ffd_data->frame_vector[i], sizeof(ndd_float8),
			  header->frame_dimension, fp) == header->frame_dimension,
		   NDD_ERROR_FILE_WRITE_FAILED);
    }

    NDD_RETURN;
}

ndd_result
ndd_ffd_data_print_to_file(const ndd_ffd_data *ffd_data, const ndd_file_header *header, FILE *fp)
{
    NDD_START;

    ndd_uint32 elements_per_frame = 0;
    NDD_VERIFY_RESULT(ndd_file_header_get_elements_per_frame(header, &elements_per_frame));
    
    /*  ե졼ǡ񤭹  */
    ndd_uint32 i, j, k;
    fputs("FFD data:\n", fp);
    NDD_FOR(i = 0 ; i < header->frame_count; i++){
	fprintf(fp, "  Vector %lu\n", i);
	NDD_FOR(j = 0; j < header->frame_dimension; j += 8){
	    fputs("   ", fp);
	    NDD_FOR(k = 0; k < 8 && j + k < header->frame_dimension; k++){
		fprintf(fp, "%7.4f, ", ffd_data->frame_vector[i][j + k]);
	    }
	    fputs("\n", fp);
	}
	fputs("\n", fp);
    }

    NDD_RETURN;
}

ndd_result
ndd_ffd_data_copy(ndd_ffd_data *dst_ffd_data, const ndd_ffd_data *src_ffd_data, const ndd_file_header *header)
{
    NDD_START;
    NDD_VERIFY_POINTER(dst_ffd_data);
    NDD_VERIFY_POINTER(src_ffd_data);
    NDD_VERIFY_POINTER(header);

    ndd_uint32 frame_count     = 0;
    ndd_uint32 frame_dimension = 0;
    NDD_VERIFY_RESULT(ndd_file_header_get_frame_count       (header, &frame_count));
    NDD_IFOK{
	frame_dimension = header->frame_dimension;
    }

    ndd_float8 **new_frame_vector = NULL;
    NDD_VERIFY_RESULT(ndd_calloc(&new_frame_vector, sizeof(ndd_float8 *) * frame_count));

    ndd_uint32 i;
    NDD_FOR(i = 0; i < frame_count; i++){
	ndd_float8 *f = NULL;
	NDD_VERIFY_RESULT(ndd_calloc(&f, sizeof(ndd_float8) * frame_dimension));
	NDD_IFOK{
	    memcpy(f, src_ffd_data->frame_vector[i], sizeof(ndd_float8) * frame_dimension);
	}

	NDD_IFOK{
	    new_frame_vector[i] = f;
	    f = NULL;
	}
	if(f != NULL){  ndd_free(&f);  }
    }

    NDD_IFOK{
	dst_ffd_data->frame_vector = new_frame_vector;
	new_frame_vector = NULL;
    }

    if(new_frame_vector != NULL){
	for(i = 0; i < frame_count; i++){ ndd_free(&new_frame_vector[i]);  }
	ndd_free(&new_frame_vector);
    }

    NDD_RETURN;
}

/*  end of file  */
