/*
    Copyright 2006-2009 Nagoya University

    This file is part of NDD (Near Duplicate Detector).

    NDD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    NDD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NDD.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ndd/ffs.h>
#include <ndd/error.h>
#include <ndd/memory.h>
#include <ndd/fvd.h>
#include <ndd/file.h>

#include <string.h>

ndd_result
ndd_ffs_create(ndd_ffs **ffs)
{ 
    NDD_START;
    NDD_VERIFY_POINTER(ffs);

    /*  ΰ  */
    ndd_ffs *new_ffs = NULL;
    NDD_VERIFY_RESULT(ndd_calloc(&new_ffs, sizeof(ndd_ffs)));

    /*  ѿν  */
    NDD_VERIFY_RESULT(ndd_file_header_create(&new_ffs->header));
    NDD_VERIFY_RESULT(ndd_ffs_data_create   (&new_ffs->ffs));

    /*  ̤Ǽ  */
    NDD_IFOK{
	*ffs = new_ffs;
	new_ffs = NULL;
    }

    /*  פʥǡդ  */
    if(new_ffs != NULL){
	ndd_ffs_destroy(&new_ffs);
    }

    NDD_RETURN;
}

ndd_result
ndd_ffs_destroy(ndd_ffs **ffs)
{
    NDD_START;
    NDD_VERIFY_POINTER(ffs);

    ndd_ffs *old_ffs;
    NDD_IFOK{
	old_ffs = *ffs;
    }

    NDD_VERIFY_RESULT(ndd_ffs_data_destroy   (&old_ffs->ffs, old_ffs->header));
    NDD_VERIFY_RESULT(ndd_file_header_destroy(&old_ffs->header));
    NDD_VERIFY_RESULT(ndd_free(&old_ffs));
    
    NDD_IFOK{
	*ffs = NULL;
    }

    NDD_RETURN;
}

ndd_result
ndd_ffs_read_from_file(ndd_ffs *ffs, FILE *fp)
{
    NDD_START;
    NDD_VERIFY_POINTER(ffs);
    NDD_VERIFY_POINTER(fp);

    /*  إåɤ߹  */
    ndd_file_header *new_header = NULL;
    NDD_VERIFY_RESULT(ndd_file_header_create(&new_header));
    NDD_VERIFY_RESULT(ndd_file_header_read_from_file(new_header, fp));

    /*  FFS ǡɤ߹  */
    ndd_ffs_data *new_ffs_data = NULL;
    NDD_VERIFY_RESULT(ndd_ffs_data_create(&new_ffs_data));
    NDD_VERIFY_RESULT(ndd_ffs_data_read_from_file(new_ffs_data, new_header, fp));

    /*  ǡǼ  */
    NDD_IFOK{
	if(ffs->ffs != NULL){
	    NDD_VERIFY_RESULT(ndd_ffs_data_destroy(&ffs->ffs, ffs->header));
	}
	if(ffs->header != NULL){
	    NDD_VERIFY_RESULT(ndd_file_header_destroy(&ffs->header));
	}
    }
    NDD_IFOK{
	ffs->header = new_header;
	ffs->ffs    = new_ffs_data;
	new_header   = NULL;
	new_ffs_data = NULL;
    }

    if(new_ffs_data != NULL){  ndd_ffs_data_destroy   (&new_ffs_data, new_header);  }
    if(new_header   != NULL){  ndd_file_header_destroy(&new_header);    }

    NDD_RETURN;
}

ndd_result
ndd_ffs_write_to_file(const ndd_ffs *ffs, FILE *fp)
{
    NDD_START;
    NDD_VERIFY_POINTER(ffs);
    NDD_VERIFY_POINTER(fp);

    /*  եإå񤭹  */
    NDD_VERIFY_RESULT(ndd_file_header_write_to_file(ffs->header, fp));
    
    /*  FFS ǡ񤭹  */
    NDD_VERIFY_RESULT(ndd_ffs_data_write_to_file(ffs->ffs, ffs->header, fp));

    NDD_RETURN;
}

ndd_result
ndd_ffs_print_to_file(const ndd_ffs *ffs, FILE *fp)
{
    NDD_START;
    NDD_VERIFY_POINTER(ffs);
    NDD_VERIFY_POINTER(fp);

    /*  եإå񤭹  */
    NDD_VERIFY_RESULT(ndd_file_header_print_to_file(ffs->header, fp));
    
    /*  FFS ǡ񤭹  */
    NDD_VERIFY_RESULT(ndd_ffs_data_print_to_file(ffs->ffs, ffs->header, fp));

    NDD_RETURN;
}

ndd_result
ndd_ffs_compare_format(const ndd_ffs *ffs1, const ndd_ffs *ffs2, ndd_bool *is_same)
{
    NDD_START;
    NDD_VERIFY_POINTER(ffs1);
    NDD_VERIFY_POINTER(ffs2);
    NDD_VERIFY_POINTER(is_same);
    
    NDD_VERIFY_RESULT(ndd_ffs_header_compare_format(ffs1->header, ffs2->header, is_same));
    
    NDD_RETURN;
}

ndd_result
ndd_ffs_header_compare_format(const ndd_file_header *header1, const ndd_file_header *header2, ndd_bool *is_same)
{
    NDD_START;
    NDD_VERIFY_POINTER(header1);
    NDD_VERIFY_POINTER(header2);
    NDD_VERIFY_POINTER(is_same);

    ndd_bool same = false;
    NDD_VERIFY_RESULT(ndd_fvd_header_compare_format(header1, header2, &same));
    NDD_IFOK{
	if(header1->frame_dimension != header2->frame_dimension){
	    NDD_SET_ERROR(NDD_ERROR_FRAME_DIMENSION_MISMATCH);
	}else if(header1->frame_sample_count != header2->frame_sample_count){
	    NDD_SET_ERROR(NDD_ERROR_FRAME_SAMPLE_COUNT_MISMATCH);
	}else{
	    same = true;
	}
    }

    NDD_IFOK{
	*is_same = same;
    }

    NDD_RETURN;
}

ndd_result
ndd_ffs_data_compare(const ndd_ffs_data *ffs_data1, const ndd_file_header *header1,
		     const ndd_ffs_data *ffs_data2, const ndd_file_header *header2, ndd_bool *is_same)
{
    NDD_START;
    NDD_VERIFY_POINTER(ffs_data1);
    NDD_VERIFY_POINTER(ffs_data2);
    NDD_VERIFY_POINTER(is_same);

    ndd_bool same = true;
    NDD_VERIFY_RESULT(ndd_ffs_header_compare_format(header1, header2, &same));

    ndd_uint32 elements_per_frame = 0;
    NDD_VERIFY_RESULT(ndd_file_header_get_elements_per_frame(header1, &elements_per_frame));
    
    ndd_uint32 i;
    NDD_FOR(i = 0; same && i < header1->frame_dimension; i++){
	if(memcmp(ffs_data1->selector_vector[i], ffs_data2->selector_vector[i], sizeof(ndd_float8) * elements_per_frame) != 0){
	    same = false;
	    NDD_SET_ERROR(NDD_ERROR_FRAME_SELECTOR_MISMATCH);
	}
    }

    NDD_IFOK{
	*is_same = same;
    }

    NDD_RETURN;
}

ndd_result
ndd_ffs_data_create(ndd_ffs_data **ffs_data)
{ 
    NDD_START;
    NDD_VERIFY_POINTER(ffs_data);

    /*  ΰ  */
    ndd_ffs_data *new_ffs_data = NULL;
    NDD_VERIFY_RESULT(ndd_calloc(&new_ffs_data, sizeof(ndd_ffs_data)));

    /*  ѿν  */

    /*  ̤Ǽ  */
    NDD_IFOK{
	*ffs_data = new_ffs_data;
	new_ffs_data = NULL;
    }

    /*  פʥǡդ  */
    if(new_ffs_data != NULL){
	ndd_free(&new_ffs_data);
    }

    NDD_RETURN;
}

ndd_result
ndd_ffs_data_destroy(ndd_ffs_data **ffs_data, const ndd_file_header *header)
{
    NDD_START;
    NDD_VERIFY_POINTER(ffs_data);
    NDD_VERIFY_POINTER(header);

    ndd_ffs_data *old_ffs_data;
    NDD_IFOK{
	old_ffs_data = *ffs_data;
    }

    ndd_uint32 i;
    NDD_FOR(i = 0; i < header->frame_dimension; i++){
	NDD_VERIFY_RESULT(ndd_free(&old_ffs_data->selector_vector[i]));
    }
    NDD_VERIFY_RESULT(ndd_free(&old_ffs_data->selector_vector));
    NDD_VERIFY_RESULT(ndd_free(&old_ffs_data));
    
    NDD_IFOK{
	*ffs_data = NULL;
    }

    NDD_RETURN;
}

ndd_result
ndd_ffs_data_read_from_file(ndd_ffs_data *ffs_data, const ndd_file_header *header, FILE *fp)
{
    NDD_START;

    /*  ե졼礭ʤɤ  */
    ndd_uint32 elements_per_frame = 0;
    NDD_VERIFY_RESULT(ndd_file_header_get_elements_per_frame(header, &elements_per_frame));

    /*  ɤ߹ ffs ǡǼΰ  */
    ndd_float8 **new_selector_vector = NULL;
    NDD_VERIFY_RESULT(ndd_calloc(&new_selector_vector, sizeof(ndd_float8 *) * header->frame_dimension));

    /*  ƥ٥ȥΥǡɤ߹  */
    ndd_uint32 i;
    NDD_FOR(i = 0 ; i < header->frame_dimension; i++){

	ndd_float8 *selector_vector = NULL;
	NDD_VERIFY_RESULT(ndd_malloc(&selector_vector, sizeof(ndd_float8) * elements_per_frame));

	NDD_VERIFY(fread(selector_vector, sizeof(ndd_float8), elements_per_frame, fp) == elements_per_frame,
		   NDD_ERROR_FILE_READ_FAILED);
	
	NDD_IFOK{
	    new_selector_vector[i] = selector_vector;
	    selector_vector = NULL;
	}

	if(selector_vector != NULL){
	    ndd_free(&selector_vector);
	}	    
    }

    /*  줿ǡѿ˳Ǽ  */
    NDD_IFOK{
	if(ffs_data->selector_vector != NULL){
	    NDD_FOR(i = 0 ; i < header->frame_dimension; i++){
		NDD_VERIFY_RESULT(ndd_free(&ffs_data->selector_vector[i]));
	    }
	    NDD_VERIFY_RESULT(ndd_free(&ffs_data->selector_vector));
	}
    }
    NDD_IFOK{
	ffs_data->selector_vector = new_selector_vector;
	new_selector_vector = NULL;
    }

    /*  Իʤɤθդ  */
    if(new_selector_vector != NULL){
	for(i = 0 ; i < header->frame_dimension; i++){
	    ndd_free(&new_selector_vector[i]);
	}
	ndd_free(&new_selector_vector);
    }

    NDD_RETURN;
}

ndd_result
ndd_ffs_data_write_to_file(const ndd_ffs_data *ffs_data, const ndd_file_header *header, FILE *fp)
{
    NDD_START;

    ndd_uint32 elements_per_frame = 0;
    NDD_VERIFY_RESULT(ndd_file_header_get_elements_per_frame(header, &elements_per_frame));
    
    /*  ե졼ǡ񤭹  */
    ndd_uint32 i;
    NDD_FOR(i = 0 ; i < header->frame_dimension; i++){
	NDD_VERIFY(fwrite(ffs_data->selector_vector[i], sizeof(ndd_float8), elements_per_frame, fp) == elements_per_frame,
		   NDD_ERROR_FILE_WRITE_FAILED);
    }

    NDD_RETURN;
}

ndd_result
ndd_ffs_data_print_to_file(const ndd_ffs_data *ffs_data, const ndd_file_header *header, FILE *fp)
{
    NDD_START;

    ndd_uint32 elements_per_frame = 0;
    NDD_VERIFY_RESULT(ndd_file_header_get_elements_per_frame(header, &elements_per_frame));
    
    /*  ե졼ǡ񤭹  */
    ndd_uint32 i, j, k;
    fputs("FFS data:\n", fp);
    NDD_FOR(i = 0 ; i < header->frame_dimension; i++){
	fprintf(fp, "  Vector %lu\n", i);
	NDD_FOR(j = 0; j < elements_per_frame; j += 8){
	    fputs("   ", fp);
	    NDD_FOR(k = 0; k < 8 && j + k < elements_per_frame; k++){
		fprintf(fp, "%7.4f, ", ffs_data->selector_vector[i][j + k]);
	    }
	    fputs("\n", fp);
	}
	fputs("\n", fp);
    }

    NDD_RETURN;
}

ndd_result
ndd_ffs_data_copy(ndd_ffs_data *dst_ffs_data, const ndd_ffs_data *src_ffs_data, const ndd_file_header *header)
{
    NDD_START;
    NDD_VERIFY_POINTER(dst_ffs_data);
    NDD_VERIFY_POINTER(src_ffs_data);
    NDD_VERIFY_POINTER(header);

    ndd_uint32 elements_per_frame = 0;
    NDD_VERIFY_RESULT(ndd_file_header_get_elements_per_frame(header, &elements_per_frame));

    ndd_float8 **new_selector_vectors      = NULL;
    ndd_uint32   new_selector_vector_count = 0;

    ndd_uint32 i;
    NDD_FOR(i = 0; i < header->frame_dimension; i++){
	ndd_float8 *new_selector_vector = NULL;
	NDD_VERIFY_RESULT(ndd_malloc(&new_selector_vector, sizeof(ndd_float8) * elements_per_frame));
	NDD_IFOK{
	    memcpy(new_selector_vector, src_ffs_data->selector_vector[i],sizeof(ndd_float8) * elements_per_frame);
	}

	NDD_VERIFY_RESULT(ndd_realloc(&new_selector_vectors, sizeof(ndd_float8 *) * (new_selector_vector_count + 1)));
	NDD_IFOK{
	    new_selector_vectors[new_selector_vector_count++] = new_selector_vector;
	    new_selector_vector = NULL;
	}

	if(new_selector_vector != NULL){  ndd_free(&new_selector_vector);  }
    }

    NDD_IFOK{
	dst_ffs_data->selector_vector = new_selector_vectors;
	new_selector_vectors = NULL;
    }

    if(new_selector_vectors != NULL){
	for(i = 0; i < new_selector_vector_count; i++){
	    ndd_free(&new_selector_vectors[i]);
	}
	ndd_free(&new_selector_vectors);
    }

    NDD_RETURN;
}

/*  end of file  */
