/*
    Copyright 2006-2009 Nagoya University

    This file is part of NDD (Near Duplicate Detector).

    NDD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    NDD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NDD.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ndd/matrix.h>
#include <ndd/error.h>
#include <ndd/memory.h>

/*    */

ndd_result
ndd_matrix_create(ndd_matrix **matrix, ndd_uint32 column_count, ndd_uint32 row_count)
{
    NDD_START;
    NDD_VERIFY_POINTER(matrix);

    ndd_matrix  *new_matrix      = NULL;
    ndd_float8 **new_matrix_data = NULL;
    NDD_VERIFY_RESULT(ndd_calloc(&new_matrix, sizeof(ndd_matrix)));
    NDD_VERIFY_RESULT(ndd_calloc(&new_matrix_data, sizeof(ndd_float8 *) * column_count));

    ndd_uint32 i;
    NDD_FOR(i = 0; i < column_count; i++){
	NDD_VERIFY_RESULT(ndd_calloc(&new_matrix_data[i], sizeof(ndd_float8) * row_count));
    }

    NDD_IFOK{
	new_matrix->data         = new_matrix_data;
	new_matrix->row_count    = row_count;
	new_matrix->column_count = column_count;
	*matrix = new_matrix;
	new_matrix      = NULL;
	new_matrix_data = NULL;
    }

    if(new_matrix_data != NULL){
	for(i = 0; i < column_count; i++){
	    if(new_matrix_data[i] != NULL){
		ndd_free(&new_matrix_data);
	    }
	}
	ndd_free(&new_matrix_data);
    }
    if(new_matrix != NULL){
	ndd_free(&new_matrix);
    }

    NDD_RETURN;
}

ndd_result
ndd_matrix_destroy(ndd_matrix **p)
{
    NDD_START;
    NDD_VERIFY_POINTER(p);

    ndd_matrix *matrix = NULL;
    NDD_IFOK{
	matrix = *p;
    }

    ndd_uint32 i;
    NDD_FOR(i = 0; i < matrix->column_count; i++){
	NDD_VERIFY_RESULT(ndd_free(&matrix->data[i]));
    }

    NDD_VERIFY_RESULT(ndd_free(&matrix->data));
    NDD_VERIFY_RESULT(ndd_free(p));
    
    NDD_RETURN;
}

ndd_result
ndd_matrix_set_column_data(ndd_matrix *matrix, ndd_uint32 column_index, const ndd_float8 *data, ndd_uint32 count)
{
    NDD_START;
    NDD_VERIFY_POINTER(matrix);
    NDD_VERIFY_POINTER(data);

    NDD_IFOK{
	if(column_index >= matrix->column_count){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}else if(count != matrix->row_count){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}
    }

    NDD_IFOK{
	memcpy(matrix->data[column_index], data, sizeof(ndd_float8) * count);
    }

    NDD_RETURN;
}

ndd_result
ndd_matrix_set_column_row_data(ndd_matrix *matrix, ndd_uint32 column_index, ndd_uint32 row_index, const ndd_float8 *data, ndd_uint32 count)
{
    NDD_START;
    NDD_VERIFY_POINTER(matrix);
    NDD_VERIFY_POINTER(data);

    NDD_IFOK{
	if(column_index >= matrix->column_count){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}else if(row_index + count > matrix->row_count){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}
    }

    NDD_IFOK{
	memcpy(matrix->data[column_index] + row_index, data, sizeof(ndd_float8) * count);
    }

    NDD_RETURN;
}

ndd_result
ndd_matrix_get_column_data(const ndd_matrix *matrix, ndd_uint32 column_index, ndd_float8 *data, ndd_uint32 count)
{
    NDD_START;
    NDD_VERIFY_POINTER(matrix);
    NDD_VERIFY_POINTER(data);

    NDD_IFOK{
	if(column_index >= matrix->column_count){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}else if(count != matrix->row_count){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}
    }

    NDD_IFOK{
	memcpy(data, matrix->data[column_index], sizeof(ndd_float8) * count);
    }

    NDD_RETURN;
}

ndd_result
ndd_matrix_swap_columns(ndd_matrix *matrix, ndd_uint32 column_index_1, ndd_uint32 column_index_2)
{
    NDD_START;
    NDD_VERIFY_POINTER(matrix);

    NDD_IFOK{
	if(column_index_1 >= matrix->column_count ||
	   column_index_2 >= matrix->column_count){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}
    }

    NDD_IFOK{
	ndd_float8 *t = matrix->data[column_index_1];
	matrix->data[column_index_1] = matrix->data[column_index_2];
	matrix->data[column_index_2] = t;
    }

    NDD_RETURN;
}

/*  end of file  */
