/*
    Copyright 2006-2009 Nagoya University

    This file is part of NDD (Near Duplicate Detector).

    NDD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    NDD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NDD.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ndd/mkffs.h>
#include <ndd/error.h>
#include <ndd/memory.h>
#include <ndd/fvd.h>
#include <ndd/ffs.h>
#include <ndd/file.h>
#include <ndd/matrix.h>
#include <ndd/vector.h>
#include <ndd/pca.h>

#include <math.h>

static int
compare_uint32(const void *x, const void *y)
{
    const ndd_uint32 v1 = *(const ndd_uint32 *)x;
    const ndd_uint32 v2 = *(const ndd_uint32 *)y;
    if(v1 > v2){
	return 1;
    }else if(v1 < v2){
	return -1;
    }else{
	return 0;
    }
}

ndd_result
ndd_create_ffs_from_fvd(ndd_ffs *dst_ffs, ndd_fvd *src_fvd, ndd_callback callback, void *user_data)
{
    NDD_START;
    NDD_VERIFY_POINTER(dst_ffs);
    NDD_VERIFY_POINTER(src_fvd);

    /*  Retrieve frame size  */
    ndd_uint32 elements_per_frame = 0;
    ndd_uint32 fvd_frame_count    = 0;
    ndd_uint32 frame_sample_count = 0;
    NDD_VERIFY_RESULT(ndd_file_header_get_elements_per_frame(src_fvd->header, &elements_per_frame));
    NDD_VERIFY_RESULT(ndd_file_header_get_frame_count       (src_fvd->header, &fvd_frame_count));
    NDD_IFOK{
	/*  Check if Kf >= mf  */
	if(dst_ffs->header->frame_sample_count < elements_per_frame){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}else if(fvd_frame_count < dst_ffs->header->frame_sample_count){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}else{
	    frame_sample_count = dst_ffs->header->frame_sample_count;
	}
    }

    /*  ξ¤򼨤ͤ  */
    ndd_uint32 ffs_dimension_limit  = 0;
    ndd_float8 ffs_proportion_limit = 0;
    NDD_IFOK{
	if(dst_ffs->header->frame_dimension > 0 && dst_ffs->header->frame_proportion > 0){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}else if(dst_ffs->header->frame_dimension == 0 && dst_ffs->header->frame_proportion <= 0){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}else{
	    ffs_dimension_limit  = dst_ffs->header->frame_dimension;
	    ffs_proportion_limit = dst_ffs->header->frame_proportion;
	}
    }

    /*
    fprintf(stdout, "elements_per_frame = %lu\n", elements_per_frame);
    fprintf(stdout, "fvd_frame_count    = %lu\n", fvd_frame_count);
    fprintf(stdout, "frame_sample_count = %lu\n", frame_sample_count);
    */

    /*  Prepare PCA source data  */
    ndd_matrix *matrix_F = NULL;
    NDD_VERIFY_RESULT(ndd_matrix_create(&matrix_F, frame_sample_count, elements_per_frame));

    ndd_uint32 *frame_indices = NULL;
    NDD_VERIFY_RESULT(ndd_malloc(&frame_indices, sizeof(ndd_uint32) * frame_sample_count));

    ndd_float8 *frame_buffer = NULL;
    NDD_VERIFY_RESULT(ndd_malloc(&frame_buffer, sizeof(ndd_float8) * elements_per_frame));

    ndd_uint32 i;
    NDD_FOR(i = 0; i < frame_sample_count; i++){
	/*  Choose i-th sample frame  */
	ndd_uint32 frame_index = (ndd_uint32)((ndd_float8)rand() * fvd_frame_count / RAND_MAX);
	frame_indices[i] = frame_index;
    }
    NDD_IFOK{
	qsort(frame_indices, frame_sample_count, sizeof(ndd_uint32), compare_uint32);
    }
    NDD_FOR(i = 0; i < frame_sample_count; i++){
	NDD_IFOK{
	    if(callback != NULL){
		NDD_VERIFY_RESULT(callback("mkffs", "create sample frame data", i, frame_sample_count, user_data));
	    }
	}
	NDD_VERIFY_RESULT(ndd_fvd_read_frame_data(src_fvd, frame_indices[i], frame_buffer, elements_per_frame));
	NDD_VERIFY_RESULT(ndd_matrix_set_column_data(matrix_F, i, frame_buffer, elements_per_frame));
    }
    NDD_IFOK{
	if(callback != NULL){
	    NDD_VERIFY_RESULT(callback("mkffs", NULL, 0, 0, user_data));
	}
    }

    if(frame_indices != NULL){ ndd_free(&frame_indices); }
    if(frame_buffer != NULL){ ndd_free(&frame_buffer); }

    /*  Execute PCA  */
    ndd_matrix *matrix_e = NULL;
    ndd_vector *vector_l = NULL;
    NDD_VERIFY_RESULT(ndd_matrix_create(&matrix_e, elements_per_frame, elements_per_frame));
    NDD_VERIFY_RESULT(ndd_vector_create(&vector_l, elements_per_frame));

    ndd_pca *pca = NULL;
    NDD_VERIFY_RESULT(ndd_pca_create(&pca));
    NDD_VERIFY_RESULT(ndd_pca_analyze(pca, matrix_e, vector_l, matrix_F, callback, user_data));
    NDD_VERIFY_RESULT(ndd_pca_destroy(&pca));

    /*  ͭͤ¤׻  */
    ndd_float8 sum_all_lambda = 0;
    NDD_FOR(i = 0; i < elements_per_frame; i++){
	sum_all_lambda += vector_l->data[i];
    }
    ndd_float8 ffs_lambda_limit = 0;
    NDD_IFOK{
	ffs_lambda_limit = sum_all_lambda * ffs_proportion_limit;
    }

    /*  ͿΨͤ˽äơͭ٥ȥ  */
    ndd_float8 sum_lambda = 0;
    ndd_float8 **new_vector_data  = NULL;
    ndd_uint32   new_vector_count = 0;
    NDD_FOR(i = 0; i < elements_per_frame; i++){
	if(ffs_dimension_limit > 0 && i >= ffs_dimension_limit){
	    break;
	}
	sum_lambda += vector_l->data[i];
	if(ffs_lambda_limit > 0 && sum_lambda > ffs_lambda_limit){
	    break;
	}

	ndd_float8 *new_vector = NULL;
	NDD_VERIFY_RESULT(ndd_malloc(&new_vector, sizeof(ndd_float8) * elements_per_frame));
	NDD_VERIFY_RESULT(ndd_matrix_get_column_data(matrix_e, i, new_vector, elements_per_frame));

	NDD_VERIFY_RESULT(ndd_realloc(&new_vector_data, sizeof(ndd_float8 *) * (new_vector_count + 1)));
	NDD_IFOK{
	    new_vector_data[new_vector_count++] = new_vector;
	    new_vector = NULL;
	}

	if(new_vector != NULL){  ndd_free(&new_vector);  }
    }
    ndd_uint32 ffs_dimension  = 0;
    ndd_float8 ffs_proportion = 0;
    NDD_IFOK{
	ffs_dimension  = new_vector_count;
	ffs_proportion = sum_lambda / sum_all_lambda;
    }

    /*  ͤǼ  */
    NDD_VERIFY_RESULT(ndd_file_header_copy(dst_ffs->header, src_fvd->header));
    NDD_IFOK{
	dst_ffs->header->marker             = NDD_FILE_MARKER_FFS;
	dst_ffs->header->version            = NDD_FILE_VERSION_FFS;
	dst_ffs->header->frame_sample_count = frame_sample_count;
	dst_ffs->header->frame_dimension    = ffs_dimension;
	dst_ffs->header->frame_proportion   = ffs_proportion;
	dst_ffs->ffs->selector_vector       = new_vector_data;
	new_vector_data = NULL;
    }

    /*  դ  */
    if(new_vector_data != NULL){
	for(i = 0; i < new_vector_count; i++){
	    ndd_free(&new_vector_data[i]);
	}
	ndd_free(&new_vector_data);
    }
    if(pca != NULL){  ndd_pca_destroy(&pca);  }
    if(matrix_F != NULL){  ndd_matrix_destroy(&matrix_F);  }
    if(matrix_e != NULL){  ndd_matrix_destroy(&matrix_e);  }
    if(vector_l != NULL){  ndd_vector_destroy(&vector_l);  }

    NDD_RETURN;
}

/*  end of file  */
