/*
    Copyright 2006-2009 Nagoya University

    This file is part of NDD (Near Duplicate Detector).

    NDD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    NDD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NDD.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ndd/mkfvd.h>
#include <ndd/error.h>
#include <ndd/memory.h>
#include <ndd/fvd.h>
#include <ndd/file.h>

#include <math.h>

#if defined(MPEGLIB_mpegscan)
#include "openmpeg.h"
#elif defined(MPEGLIB_libmpeg3)
#include "libmpeg3.h"
#else
#error No MPEGLIB specified.
#endif

ndd_result
ndd_create_fvd_from_mpeg(ndd_fvd *dst_fvd, const char *mpeg_file,
			 ndd_callback callback, void *user_data)
{
    NDD_START;
    NDD_VERIFY_POINTER(dst_fvd);
    NDD_VERIFY_POINTER(mpeg_file);

    /*  Open MPEG file  */
    ndd_float8 mpeg_frames_per_second = 0;
    ndd_uint32 mpeg_width  = 0;
    ndd_uint32 mpeg_height = 0;

#if defined(MPEGLIB_libmpeg3)
    mpeg3_t *file = NULL;
    NDD_VERIFY((file = mpeg3_open((char *)mpeg_file)) != NULL, NDD_ERROR_MPEG_FILE_OPEN_FAILED);

    NDD_IFOK{
	if(!mpeg3_has_video(file)){
	    NDD_SET_ERROR(NDD_ERROR_MPEG_HAS_NO_VIDEO_STREAM);
	}else if(mpeg3_total_vstreams(file) < 1){
	    NDD_SET_ERROR(NDD_ERROR_MPEG_HAS_NO_VIDEO_STREAM);
	}
    }
    NDD_IFOK{
	mpeg_frames_per_second = mpeg3_frame_rate(file, 0);
	mpeg_width  = mpeg3_video_width (file, 0);
	mpeg_height = mpeg3_video_height(file, 0);
	fprintf(stdout, "%ld\n", mpeg3_video_frames(file, 0));
    }
#else
    int *values = NULL;
    NDD_VERIFY((values = OpenMpeg2((char *)mpeg_file)) != NULL, NDD_ERROR_MPEG_FILE_OPEN_FAILED);
    NDD_IFOK{
	mpeg_frames_per_second = 29.97;
	mpeg_width  = values[0];
	mpeg_height = values[1];
    }
#endif

    /*  Retrieve fvd frame_size  */
    ndd_uint32 fvd_region_width  = 0;
    ndd_uint32 fvd_region_height = 0;
    NDD_IFOK{
	fvd_region_width  = dst_fvd->header->region_width;
	fvd_region_height = dst_fvd->header->region_height;
	if(fvd_region_width == 0){
	    fvd_region_width = mpeg_width;
	}
	if(fvd_region_height == 0){
	    fvd_region_height = mpeg_height;
	}
    }

    ndd_uint32 fvd_region_left   = 0;
    ndd_uint32 fvd_region_top    = 0;
    NDD_IFOK{
	if(mpeg_width  < dst_fvd->header->region_left + fvd_region_width ||
	   mpeg_height < dst_fvd->header->region_top  + fvd_region_height){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}
	fvd_region_left = dst_fvd->header->region_left;
	fvd_region_top  = dst_fvd->header->region_top;
    }

    ndd_uint32 fvd_block_width  = 0;
    ndd_uint32 fvd_block_height = 0;
    NDD_IFOK{
	if(dst_fvd->header->block_width == 0 || dst_fvd->header->block_height == 0){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}else{
	    if(fvd_region_width  % dst_fvd->header->block_width  != 0 ||
	       fvd_region_height % dst_fvd->header->block_height != 0){
		fvd_region_width  = fvd_region_width  / dst_fvd->header->block_width  * dst_fvd->header->block_width;
		fvd_region_height = fvd_region_height / dst_fvd->header->block_height * dst_fvd->header->block_height;
	    }
	    fvd_block_width  = dst_fvd->header->block_width;
	    fvd_block_height = dst_fvd->header->block_height;
	}
    }

    ndd_uint32 fvd_width  = 0;
    ndd_uint32 fvd_height = 0;
    NDD_IFOK{
	fvd_width  = fvd_region_height / fvd_block_height;
	fvd_height = fvd_region_width  / fvd_block_width;
	if(fvd_width == 0 || fvd_height == 0){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}
    }

    ndd_uint32 fvd_color_resolution = 0;
    NDD_IFOK{
	fvd_color_resolution = dst_fvd->header->color_resolution;
    }

    NDD_IFOK{
	dst_fvd->header->marker            = NDD_FILE_MARKER_FVD;
	dst_fvd->header->version           = NDD_FILE_VERSION_FVD;
	dst_fvd->header->screen_width      = mpeg_width;
	dst_fvd->header->screen_height     = mpeg_height;
	dst_fvd->header->frames_per_second = mpeg_frames_per_second;
	dst_fvd->header->region_width      = fvd_region_width;
	dst_fvd->header->region_height     = fvd_region_height;
    }

    /*  New frame data  */
//    ndd_float8 **new_frame_data  = NULL;
    ndd_uint32   new_frame_count = 0;

    ndd_float8 *fij_buffer_r = NULL;
    ndd_float8 *fij_buffer_g = NULL;
    ndd_float8 *fij_buffer_b = NULL;
    NDD_VERIFY_RESULT(ndd_malloc(&fij_buffer_r, fvd_width * fvd_height * sizeof(ndd_float8)));
    NDD_VERIFY_RESULT(ndd_malloc(&fij_buffer_g, fvd_width * fvd_height * sizeof(ndd_float8)));
    NDD_VERIFY_RESULT(ndd_malloc(&fij_buffer_b, fvd_width * fvd_height * sizeof(ndd_float8)));

    /*  Read each frame from MPEG file  */
    ndd_uint32  frame_index = 0;

#if defined(MPEGLIB_libmpeg3)
    ndd_uint8 **rgb = NULL;
    NDD_VERIFY_RESULT(ndd_calloc(&rgb, sizeof(ndd_uint8 *) * mpeg_height));
    ndd_uint32 i;
    NDD_FOR(i = 0; i < mpeg_height; i++){
	NDD_VERIFY_RESULT(ndd_malloc(&rgb[i], sizeof(ndd_uint8) * (mpeg_width * 3 + 4)));
    }
#else
    ndd_uint8 *rgb = NULL;
#endif

#if defined(MPEGLIB_libmpeg3)
    NDD_WHILE(mpeg3_read_frame(file, rgb, 0, 0, mpeg_width, mpeg_height, mpeg_width, mpeg_height, MPEG3_RGB888, 0) == 0){
#else
    NDD_WHILE((rgb = getframe()) != NULL){
#endif
	NDD_IFOK{
	    if(callback != NULL){
		NDD_VERIFY_RESULT(callback("mkfvd", "read frame data", frame_index, 0, user_data));
	    }
	}

	/*  Prepare buffer for f_ij  */
	ndd_uint32 i;
	ndd_uint32 fvd_index = 0;
	NDD_FOR(fvd_index = 0; fvd_index < fvd_width * fvd_height; fvd_index++){
	    fij_buffer_r[fvd_index] = 0;
	    fij_buffer_g[fvd_index] = 0;
	    fij_buffer_b[fvd_index] = 0;
	}

	ndd_uint32 mpeg_x, mpeg_y, fvd_x, fvd_y;
	ndd_uint32 mpeg_x_start = fvd_region_left;
	ndd_uint32 mpeg_x_end   = fvd_region_left + fvd_region_width;
	ndd_uint32 mpeg_y_start = fvd_region_top;
	ndd_uint32 mpeg_y_end   = fvd_region_top + fvd_region_height;
	NDD_FOR(mpeg_y = mpeg_y_start; mpeg_y < mpeg_y_end; mpeg_y++){
	    fvd_y = (mpeg_y - fvd_region_top) / fvd_block_height;

	    NDD_FOR(mpeg_x = mpeg_x_start; mpeg_x < mpeg_x_end; mpeg_x++){
		fvd_x = (mpeg_x - fvd_region_left) / fvd_block_width;
	    
		ndd_uint32 fvd_index  = (fvd_y  * fvd_width  + fvd_x);

#if defined(MPEGLIB_libmpeg3)
		fij_buffer_r[fvd_index] += rgb[mpeg_y][mpeg_x * 3 + 0] * fvd_color_resolution / 256;
		fij_buffer_g[fvd_index] += rgb[mpeg_y][mpeg_x * 3 + 1] * fvd_color_resolution / 256;
		fij_buffer_b[fvd_index] += rgb[mpeg_y][mpeg_x * 3 + 2] * fvd_color_resolution / 256;
#else
		ndd_uint32 mpeg_index = (mpeg_y * mpeg_width + mpeg_x) * 3;
		fij_buffer_r[fvd_index] += rgb[mpeg_index + 0] * fvd_color_resolution / 256;
		fij_buffer_g[fvd_index] += rgb[mpeg_index + 1] * fvd_color_resolution / 256;
		fij_buffer_b[fvd_index] += rgb[mpeg_index + 2] * fvd_color_resolution / 256;
#endif
	    }
	}

	/*  Calculate mu  */
	ndd_float8 sum_r = 0, sum_g = 0, sum_b = 0;
	NDD_FOR(fvd_index = 0; fvd_index < fvd_width * fvd_height; fvd_index++){
	    sum_r += fij_buffer_r[fvd_index];
	    sum_g += fij_buffer_g[fvd_index];
	    sum_b += fij_buffer_b[fvd_index];
	}
	ndd_float8 mu = 0;
	NDD_IFOK{
	    mu = (sum_r + sum_g + sum_b) / (fvd_height * fvd_width * 3);
	}

	/*  Calculate f~  */
	NDD_FOR(fvd_index = 0; fvd_index < fvd_width * fvd_height; fvd_index++){
	    fij_buffer_r[fvd_index] -= mu;
	    fij_buffer_g[fvd_index] -= mu;
	    fij_buffer_b[fvd_index] -= mu;
	}
	
	/*  Calculate norm(f~)  */
	sum_r = sum_g = sum_b = 0;
	NDD_FOR(fvd_index = 0; fvd_index < fvd_width * fvd_height; fvd_index++){
	    sum_r += fij_buffer_r[fvd_index] * fij_buffer_r[fvd_index];
	    sum_g += fij_buffer_g[fvd_index] * fij_buffer_g[fvd_index];
	    sum_b += fij_buffer_b[fvd_index] * fij_buffer_b[fvd_index];
	}
	ndd_float8 norm = 0, inorm = 1;
	NDD_IFOK{
	    norm  = sqrt(sum_r + sum_g + sum_b);
	    inorm = 1 / norm;
	}

	/*  Calculatnew_frame_datae f^  */
	NDD_FOR(fvd_index = 0; fvd_index < fvd_width * fvd_height; fvd_index++){
	    fij_buffer_r[fvd_index] *= inorm;
	    fij_buffer_g[fvd_index] *= inorm;
	    fij_buffer_b[fvd_index] *= inorm;
	}

	/*  Store data  */
	ndd_float8 *fij_buffer = NULL;
	NDD_VERIFY_RESULT(ndd_malloc(&fij_buffer, fvd_width * fvd_height * 3 * sizeof(ndd_float8)));
	NDD_FOR(i = 0; i < fvd_width * fvd_height; i++){
	    fij_buffer[i * 3 + 0] = fij_buffer_r[i];
	    fij_buffer[i * 3 + 1] = fij_buffer_g[i];
	    fij_buffer[i * 3 + 2] = fij_buffer_b[i];
	    /*
	    if(isnan(fij_buffer[i * 3 + 0]) || isnan(fij_buffer[i * 3 + 1]) || isnan(fij_buffer[i * 3 + 2])){
		abort();
	    }
	    */
	}

	/*
	NDD_VERIFY_RESULT(ndd_realloc(&new_frame_data, sizeof(ndd_float8 *) * (new_frame_count + 1)));
	NDD_IFOK{
	    new_frame_data[new_frame_count++] = fij_buffer;
	    fij_buffer = NULL;
	}
	*/
	NDD_IFOK{
	    dst_fvd->header->total_frame_count = ++new_frame_count;
	}
	NDD_VERIFY_RESULT(ndd_fvd_write_frame_data(dst_fvd, frame_index, fij_buffer, fvd_width * fvd_height * 3));

	frame_index++;

	/*  Free stuff  */
	if(fij_buffer != NULL){  free(fij_buffer);  }
    }
    NDD_IFOK{
	if(callback != NULL){
	    NDD_VERIFY_RESULT(callback("mkfvd", NULL, 0, 0, user_data));
	}
    }

    /*  Store created fvd data  */
    //    NDD_VERIFY_RESULT(ndd_fvd_discard_frame_data(dst_fvd));
//    NDD_IFOK{
//	dst_fvd->header->marker            = NDD_FILE_MARKER_FVD;
//	dst_fvd->header->version           = NDD_FILE_VERSION_FVD;
//	dst_fvd->header->screen_width      = mpeg_width;
//	dst_fvd->header->screen_height     = mpeg_height;
//	dst_fvd->header->frames_per_second = 30;
//	dst_fvd->header->total_frame_count = new_frame_count;
//	dst_fvd->header->frame_offset      = 0;
//	dst_fvd->header->frame_count       = new_frame_count;
//	dst_fvd->fvd->frame_data           = new_frame_data;
//	new_frame_data = NULL;
//    }

    /*  Free/close stuff  */
    if(fij_buffer_r != NULL){  free(fij_buffer_r);  }
    if(fij_buffer_g != NULL){  free(fij_buffer_g);  }
    if(fij_buffer_b != NULL){  free(fij_buffer_b);  }

#if defined(MPEGLIB_libmpeg3)
    if(file != NULL){
	mpeg3_close(file);
    }
#else
    if(values != NULL){
	CloseMpeg2();
    }
#endif

//    if(new_frame_data != NULL){
//	ndd_uint32 i;
//	for(i = 0;i < new_frame_count; i++){
//	    ndd_free(&new_frame_data[i]);
//	}
//	ndd_free(&new_frame_data);
//    }
    
    NDD_RETURN;
}

/*  end of file  */
