/*
    Copyright 2006-2009 Nagoya University

    This file is part of NDD (Near Duplicate Detector).

    NDD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    NDD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NDD.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ndd/mksfs.h>
#include <ndd/error.h>
#include <ndd/memory.h>
#include <ndd/ffd.h>
#include <ndd/ffs.h>
#include <ndd/sfs.h>
#include <ndd/file.h>
#include <ndd/matrix.h>
#include <ndd/vector.h>
#include <ndd/pca.h>

#include <math.h>

ndd_result
ndd_create_sfs_from_ffd(ndd_sfs *dst_sfs, const ndd_ffd *src_ffd, ndd_callback callback, void *user_data)
{
    NDD_START;
    NDD_VERIFY_POINTER(dst_sfs);
    NDD_VERIFY_POINTER(src_ffd);

    /*  ե졼  */
    ndd_uint32 frame_count          = 0;
    ndd_uint32 segment_sample_count = 0;  /*  Kv  */
    ndd_uint32 segment_size         = 0;  /*  Nf  */
    ndd_uint32 frame_dimension      = 0;  /*  Df  */
    ndd_uint32 elements_per_segment = 0;  /*  DfNf  */
    NDD_VERIFY_RESULT(ndd_file_header_get_frame_count(src_ffd->header, &frame_count));
    NDD_IFOK{
	/*  Check if Kv >= mv  */
	if(dst_sfs->header->segment_sample_count < src_ffd->header->frame_dimension * dst_sfs->header->segment_size){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}else if(frame_count < dst_sfs->header->segment_sample_count){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}else if(frame_count < dst_sfs->header->segment_size){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}else{
	    segment_sample_count = dst_sfs->header->segment_sample_count;
	    segment_size         = dst_sfs->header->segment_size;
	    frame_dimension      = src_ffd->header->frame_dimension;
	    elements_per_segment = segment_size * frame_dimension;
	}
    }

    /*  ξ¤򼨤ͤ  */
    ndd_uint32 sfs_dimension_limit  = 0;
    ndd_float8 sfs_proportion_limit = 0;
    NDD_IFOK{
	if(dst_sfs->header->segment_dimension > 0 && dst_sfs->header->segment_proportion > 0){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}else if(dst_sfs->header->segment_dimension == 0 && dst_sfs->header->segment_proportion <= 0){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}else{
	    sfs_dimension_limit  = dst_sfs->header->segment_dimension;
	    sfs_proportion_limit = dst_sfs->header->segment_proportion;
	}
    }

    /*  PCA Ԥǡ  */
    ndd_matrix *matrix_V = NULL;
    NDD_VERIFY_RESULT(ndd_matrix_create(&matrix_V, segment_sample_count, elements_per_segment));
    
    ndd_uint32 i;
    NDD_FOR(i = 0; i < segment_sample_count; i++){
	/*  Choose i-th sample segment  */
	ndd_uint32 frame_index = (ndd_uint32)((ndd_float8)rand() * (frame_count - segment_size) / RAND_MAX);

	NDD_IFOK{
	    if(callback != NULL){
		NDD_VERIFY_RESULT(callback("mksfs", "create sample segment data", i, segment_sample_count, user_data));
	    }
	}

	ndd_uint32 j;
	NDD_FOR(j = 0; j < segment_size; j++){
	    NDD_VERIFY_RESULT(ndd_matrix_set_column_row_data(matrix_V, i, j * frame_dimension,
							     src_ffd->ffd->frame_vector[frame_index + j],
							     frame_dimension));
	}
    }
    NDD_IFOK{
	if(callback != NULL){
	    NDD_VERIFY_RESULT(callback("mksfs", NULL, 0, 0, user_data));
	}
    }

    /*  PCA ¹  */
    ndd_matrix *matrix_e = NULL;
    ndd_vector *vector_l = NULL;
    NDD_VERIFY_RESULT(ndd_matrix_create(&matrix_e, elements_per_segment, elements_per_segment));
    NDD_VERIFY_RESULT(ndd_vector_create(&vector_l, elements_per_segment));

    ndd_pca *pca = NULL;
    NDD_VERIFY_RESULT(ndd_pca_create(&pca));
    NDD_VERIFY_RESULT(ndd_pca_analyze(pca, matrix_e, vector_l, matrix_V, callback, user_data));
    NDD_VERIFY_RESULT(ndd_pca_destroy(&pca));

    /*  ͭͤ¤׻  */
    ndd_float8 sum_all_lambda = 0;
    NDD_FOR(i = 0; i < elements_per_segment; i++){
	sum_all_lambda += vector_l->data[i];
    }
    ndd_float8 sfs_lambda_limit = 0;
    NDD_IFOK{
	sfs_lambda_limit = sum_all_lambda * sfs_proportion_limit;
    }

    /*  ͿΨͤ˽äơͭ٥ȥ  */
    ndd_float8 sum_lambda = 0;
    ndd_float8 **new_vector_data  = NULL;
    ndd_uint32   new_vector_count = 0;
    NDD_FOR(i = 0; i < elements_per_segment; i++){
	if(sfs_dimension_limit > 0 && i >= sfs_dimension_limit){
	    break;
	}
	sum_lambda += vector_l->data[i];
	if(sfs_lambda_limit > 0 && sum_lambda > sfs_lambda_limit){
	    break;
	}

	ndd_float8 *new_vector = NULL;
	NDD_VERIFY_RESULT(ndd_malloc(&new_vector, sizeof(ndd_float8) * elements_per_segment));
	NDD_VERIFY_RESULT(ndd_matrix_get_column_data(matrix_e, i, new_vector, elements_per_segment));

	NDD_VERIFY_RESULT(ndd_realloc(&new_vector_data, sizeof(ndd_float8 *) * (new_vector_count + 1)));
	NDD_IFOK{
	    new_vector_data[new_vector_count++] = new_vector;
	    new_vector = NULL;
	}

	if(new_vector != NULL){  ndd_free(&new_vector);  }
    }
    ndd_uint32 sfs_dimension  = 0;
    ndd_float8 sfs_proportion = 0;
    NDD_IFOK{
	sfs_dimension  = new_vector_count;
	sfs_proportion = sum_lambda / sum_all_lambda;
    }

    /*  ͤǼ  */
    NDD_VERIFY_RESULT(ndd_file_header_copy(dst_sfs->header, src_ffd->header));
    NDD_VERIFY_RESULT(ndd_ffs_data_copy(dst_sfs->ffs, src_ffd->ffs, src_ffd->header));
    NDD_IFOK{
	dst_sfs->header->marker               = NDD_FILE_MARKER_SFS;
	dst_sfs->header->version              = NDD_FILE_VERSION_SFS;
	dst_sfs->header->segment_size         = segment_size;
	dst_sfs->header->segment_sample_count = segment_sample_count;
	dst_sfs->header->segment_dimension    = sfs_dimension;
	dst_sfs->header->segment_proportion   = sfs_proportion;
	dst_sfs->header->segments_per_second  = src_ffd->header->frames_per_second;
	dst_sfs->header->total_segment_count  = src_ffd->header->total_frame_count;
	dst_sfs->sfs->selector_vector         = new_vector_data;
	new_vector_data = NULL;
    }

    /*  դ  */
    if(new_vector_data != NULL){
	for(i = 0; i < new_vector_count; i++){
	    ndd_free(&new_vector_data[i]);
	}
	ndd_free(&new_vector_data);
    }
    if(pca != NULL){  ndd_pca_destroy(&pca);  }
    if(matrix_V != NULL){  ndd_matrix_destroy(&matrix_V);  }
    if(matrix_e != NULL){  ndd_matrix_destroy(&matrix_e);  }
    if(vector_l != NULL){  ndd_vector_destroy(&vector_l);  }

    NDD_RETURN;
}

/*  end of file  */
