/*
    Copyright 2006-2009 Nagoya University

    This file is part of NDD (Near Duplicate Detector).

    NDD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    NDD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NDD.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ndd/pca.h>
#include <ndd/error.h>
#include <ndd/memory.h>
#include <ndd/matrix.h>
#include <ndd/vector.h>

#include <math.h>

static ndd_result jacobi(ndd_matrix *a, ndd_matrix *x, ndd_callback callback, void *user_data);

ndd_result
ndd_pca_create(ndd_pca **pca)
{
    NDD_START;
    NDD_VERIFY_POINTER(pca);

    ndd_pca *new_pca = NULL;
    NDD_VERIFY_RESULT(ndd_malloc(&new_pca, sizeof(ndd_pca)));

    NDD_IFOK{
	*pca = new_pca;
	new_pca = NULL;
    }

    if(new_pca != NULL){
	ndd_free(&new_pca);
    }

    NDD_RETURN;
}

ndd_result
ndd_pca_destroy(ndd_pca **pca)
{
    NDD_START;
    NDD_VERIFY_POINTER(pca);

    NDD_VERIFY_RESULT(ndd_free(pca));
    NDD_IFOK{
	*pca = NULL;
    }

    NDD_RETURN;
}

ndd_result
ndd_pca_analyze(ndd_pca *pca,
		ndd_matrix *dst_coef_matrix,
		ndd_vector *dst_lambda_vector,
		const ndd_matrix *src_matrix,
		ndd_callback callback, void *user_data)
{
    NDD_START;
    NDD_VERIFY_POINTER(pca);
    NDD_VERIFY_POINTER(dst_coef_matrix);
    NDD_VERIFY_POINTER(dst_lambda_vector);
    NDD_VERIFY_POINTER(src_matrix);

    /*  Check & retrieve parameters  */
    ndd_uint32 element_count = 0;
    ndd_uint32 sample_count  = 0;
    NDD_IFOK{
	if(dst_coef_matrix->column_count != dst_coef_matrix->row_count ||
	   dst_coef_matrix->column_count != dst_lambda_vector->count ||
	   dst_coef_matrix->column_count != src_matrix->row_count){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}
	element_count = src_matrix->row_count;
	sample_count  = src_matrix->column_count;
    }

    /*  Create correlation matrix  */
    ndd_matrix *matrix_c = NULL;
    NDD_VERIFY_RESULT(ndd_matrix_create(&matrix_c, src_matrix->row_count, src_matrix->row_count));

    /*  Prepare mean/sigma of each element  */
    ndd_float8 *mu = NULL, *sigma = NULL;
    NDD_VERIFY_RESULT(ndd_malloc(&mu,    sizeof(ndd_float8) * element_count));
    NDD_VERIFY_RESULT(ndd_malloc(&sigma, sizeof(ndd_float8) * element_count));

    ndd_uint32 i, j;
    NDD_FOR(i = 0; i < element_count; i++){

	ndd_float8 sum = 0, ssum = 0;
	NDD_FOR(j = 0; j < sample_count; j++){
	    sum  += src_matrix->data[j][i];
	    ssum += src_matrix->data[j][i] * src_matrix->data[j][i];
	}

	mu[i]    = sum / sample_count;
	sigma[i] = sqrt(ssum / sample_count - mu[i] * mu[i]);
    }

    /*  Calculate corrlation matrix  */
    NDD_FOR(i = 0; i < element_count; i++){
	NDD_FOR(j = i + 1; j < element_count; j++){
	    matrix_c->data[i][j] = 0.0;
	}
    }

    ndd_uint32 k;
    NDD_FOR(k = 0; k < sample_count; k++){
	ndd_float8 *data = src_matrix->data[k];
	NDD_IFOK{
	    if(callback != NULL){
		NDD_VERIFY_RESULT(callback("pca", "correlation matrix", k, sample_count, user_data));
	    }
	}

	NDD_FOR(i = 0; i < element_count; i++){
	    NDD_FOR(j = i + 1; j < element_count; j++){
		matrix_c->data[i][j] += (data[i] - mu[i]) * (data[j] - mu[j]);
	    }
	}
    }
    NDD_FOR(i = 0; i < element_count; i++){
	matrix_c->data[i][i] = 1.0;

	NDD_FOR(j = i + 1; j < element_count; j++){
	    matrix_c->data[i][j] /= (sample_count * sigma[i] * sigma[j]);
	    matrix_c->data[j][i]  = matrix_c->data[i][j];
	}
    }
    NDD_IFOK{
	if(callback != NULL){
	    NDD_VERIFY_RESULT(callback("pca", NULL, 0, 0, user_data));
	}
    }

    /*  ǡåҥåȤʤΤǡ嵭ѹ
    NDD_FOR(i = 0; i < element_count; i++){
	NDD_VERIFY_RESULT(callback("pca", "correlation matrix", i, element_count, user_data));
        NDD_FOR(j = 0; j < element_count; j++){

            ndd_float8 correlation = 0;
            if(i == j){
                correlation = 1.0;
            }else if(i > j){
                correlation = matrix_c->data[j][i];
            }else{
                ndd_float8 sum = 0;
                ndd_uint32 k;
                NDD_FOR(k = 0; k < sample_count; k++){
                    sum += (src_matrix->data[k][i] - mu[i]) * (src_matrix->data[k][j] - mu[j]);
                }
                correlation = sum / sample_count / (sigma[i] * sigma[j]);
            }

            matrix_c->data[i][j] = correlation;
        }
    }
    NDD_VERIFY_RESULT(callback("pca", NULL,0, 0, user_data));
    */
    
    if(0){
	fprintf(stdout, "Source matrix:\n");
	for(i = 0; i < sample_count; i++){
	    fprintf(stdout, "Column %lu\n", i);
	    for(j = 0; j < element_count; j++){
		fprintf(stdout, "%10.5f, ", src_matrix->data[i][j]);
		if((j + 1) % 6 == 0){
		    fputs("\n", stdout);
		}

		if(isnan(src_matrix->data[i][j])){
		    fprintf(stdout, "NaN: %lu,%lu\n", i, j);
		}
	    }
	    fputs("\n", stdout);
	}
    }

    if(0){
	fprintf(stdout, "mu:\n");
	fputs("    ", stdout);
	for(j = 0; j < element_count; j++){
	    fprintf(stdout, "%10.5f, ", mu[j]);
	}
	fputs("\n", stdout);

	fprintf(stdout, "sigma:\n");
	fputs("    ", stdout);
	for(j = 0; j < element_count; j++){
	    fprintf(stdout, "%10.5f, ", sigma[j]);
	}
	fputs("\n", stdout);

	fprintf(stdout, "Correlation matrix:\n");
	for(i = 0; i < element_count; i++){
	    fputs("    ", stdout);
	    for(j = 0; j < element_count; j++){
		fprintf(stdout, "%10.5f, ", matrix_c->data[i][j]);
	    }
	    fputs("\n", stdout);
	}
    }

    NDD_VERIFY_RESULT(jacobi(matrix_c, dst_coef_matrix, callback, user_data));
    if(0){
	fprintf(stdout, "Eigen value:\n");
	fputs("    ", stdout);
	for(j = 0; j < element_count; j++){
	    fprintf(stdout, "%10.5f, ", matrix_c->data[j][j]);
	}
	fputs("\n", stdout);

	fprintf(stdout, "Eigen vector matrix:\n");
	for(i = 0; i < element_count; i++){
	    fputs("    ", stdout);
	    for(j = 0; j < element_count; j++){
		fprintf(stdout, "%10.5f, ", dst_coef_matrix->data[i][j]);
	    }
	    fputs("\n", stdout);
	}
    }

    NDD_FOR(i = 0; i < element_count; i++){
	dst_lambda_vector->data[i] = matrix_c->data[i][i];
    }

    if(matrix_c){  ndd_matrix_destroy(&matrix_c);  }
    if(mu)      {  ndd_free(&mu);  }
    if(sigma)   {  ndd_free(&sigma);  }

    NDD_RETURN;
}

/*  C ˤ륢르ꥺŵҼ  */

#define EPS         1E-6  /* Ƹ */
#define TINY        1E-20 /* 0 ȸʤƤ褤 */
#define MAX_ITER    100   /* η֤ */
#define rotate(a, i, j, k, l) {      \
    double x = a->data[i][j], y = a->data[k][l]; \
    a->data[i][j] = x * c - y * s;         \
    a->data[k][l] = x * s + y * c;         }

/*  䥳ˡˤäƼоι a θͭͭ͡٥ȥ롣
    a Ͼ񤭤졢гʬ˸ͭͤи롣
    w ϳ󤬸ͭ٥ȥȤʤ(cf. vec[i] = w->data[i])
*/
ndd_result
jacobi(ndd_matrix *a, ndd_matrix *w, ndd_callback callback, void *user_data)
{

    NDD_START;
    NDD_VERIFY_POINTER(a);
    NDD_VERIFY_POINTER(w);

    ndd_uint32 size = 0;
    NDD_IFOK{
	if(a->column_count != a->row_count ||
	   w->column_count != w->row_count ||
	   a->column_count != w->column_count){
	    NDD_SET_ERROR(NDD_ERROR_INVALID_PARAMETER);
	}else{
	    size = a->column_count;
	}
    }

    ndd_uint32 i, j, k;
    ndd_float8 t, c, s, offdiag;

    s = offdiag = 0;
    NDD_FOR(j = 0; j < size; j++){
        NDD_FOR(k = 0; k < size; k++){
	    w->data[j][k] = 0;
	}

        w->data[j][j] = 1;  s += a->data[j][j] * a->data[j][j];
        NDD_FOR(k = j + 1; k < size; k++){
            offdiag += a->data[j][k] * a->data[j][k];
	}
    }

    ndd_float8 tolerance = EPS * EPS * (s / 2 + offdiag);
    ndd_uint32 iter = 0;
    NDD_FOR(iter = 1; iter <= MAX_ITER; iter++){
	NDD_IFOK{
	    if(callback != NULL){
		NDD_VERIFY_RESULT(callback("pca", "eigen vector", iter, 0, user_data));
	    }
	}

        offdiag = 0;
        NDD_FOR(j = 0; j < size - 1; j++){
            NDD_FOR(k = j + 1; k < size; k++){
                offdiag += a->data[j][k] * a->data[j][k];
	    }
	}
	if(offdiag < tolerance){
	    break;
	}

        NDD_FOR(j = 0; j < size - 1; j++){
            NDD_FOR(k = j + 1; k < size; k++){
                if(fabs(a->data[j][k]) < TINY){
		    continue;
		}

                t = (a->data[k][k] - a->data[j][j]) / (2 * a->data[j][k]);
                if(t >= 0){
		    t = 1 / (t + sqrt(t * t + 1));
		}else{
		    t = 1 / (t - sqrt(t * t + 1));
		}
                c = 1 / sqrt(t * t + 1);
                s = t * c;  t *= a->data[j][k];
                a->data[j][j] -= t;  a->data[k][k] += t;  a->data[j][k] = 0;

                NDD_FOR(i = 0; i < j; i++){
		    rotate(a, i, j, i, k);
		}
                NDD_FOR(i = j + 1; i < k; i++){
		    rotate(a, j, i, i, k);
		}
                NDD_FOR(i = k + 1; i < size; i++){
		    rotate(a, j, i, k, i);
		}
                NDD_FOR(i = 0; i < size; i++){
		    rotate(w, j, i, k, i);
		}
            }
        }
    }
    NDD_IFOK{
	if(callback != NULL){
	    NDD_VERIFY_RESULT(callback("pca", NULL, 0, 0, user_data));
	}
    }
    if(iter > MAX_ITER){
	NDD_SET_ERROR(NDD_ERROR_NOT_CONVERGED);
    }

    NDD_FOR(i = 0; i < size - 1; i++){
        k = i;  t = a->data[k][k];
        NDD_FOR(j = i + 1; j < size; j++){
            if(a->data[j][j] > t){
		k = j;  t = a->data[k][k];
	    }
	}

        a->data[k][k] = a->data[i][i];  a->data[i][i] = t;

	NDD_VERIFY_RESULT(ndd_matrix_swap_columns(w, k, i));
    }

    NDD_RETURN;
}

/*  end of file  */
