/*
    Copyright 2006-2009 Nagoya University

    This file is part of NDD (Near Duplicate Detector).

    NDD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    NDD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NDD.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ndd/search.h>
#include <ndd/error.h>
#include <ndd/sfd.h>
#include <ndd/sfs.h>
#include <ndd/ffd.h>
#include <ndd/ffs.h>
#include <ndd/fvd.h>
#include <ndd/file.h>
#include <ndd/answer.h>
#include <ndd/memory.h>

#include <math.h>

static ndd_float8
calculate_distance(const ndd_float8 *v1,
		   const ndd_float8 *v2,
		   ndd_uint32 count,
		   ndd_float8 threshold)
{
    ndd_float8 distance = 0;
    ndd_uint32 i;
    for(i = 0; i < count; i++){
	distance += (v1[i] - v2[i]) * (v1[i] - v2[i]);
	if(distance > threshold){
	    break;
	}
    }

    return distance;
}

static int
compare_answer_distance(const void *x, const void *y)
{
    const ndd_answer *a1 = *((const ndd_answer **)x);
    const ndd_answer *a2 = *((const ndd_answer **)y);
    if(a1->distance_avg < a2->distance_avg){
	return -1;
    }
    if(a1->distance_avg > a2->distance_avg){
	return 1;
    }
    return 0;
}

static int
compare_answer_duration(const void *x, const void *y)
{
    const ndd_answer *a1 = *((const ndd_answer **)x);
    const ndd_answer *a2 = *((const ndd_answer **)y);
    if(a1->frame_count > a2->frame_count){
	return -1;
    }
    if(a1->frame_count < a2->frame_count){
	return 1;
    }
    return 0;
}

static int
print_time(ndd_float8 secs, FILE *stream)
{
    ndd_uint32 hours    = (ndd_uint32)floor(secs / 3600);
    ndd_uint32 minutes  = (ndd_uint32)fmod(floor(secs / 60),     60);
    ndd_uint32 seconds  = (ndd_uint32)fmod(floor(secs / 1),      60);
    ndd_uint32 mseconds = (ndd_uint32)fmod(floor(secs * 1000), 1000);

    fprintf(stream, "%02lu:%02lu:%02lu.%03lu", hours, minutes, seconds, mseconds);

    return 0;
}

ndd_result
ndd_search(const ndd_sfd *query_sfd,  const ndd_ffd *query_ffd,  ndd_fvd *query_fvd,
	   const ndd_sfd *stored_sfd, const ndd_ffd *stored_ffd, ndd_fvd *stored_fvd,
	   ndd_search_merge_level search_merge_level,	   
	   ndd_callback callback, void *user_data)
{
    NDD_START;
    NDD_VERIFY_POINTER(query_sfd);
    NDD_VERIFY_POINTER(query_ffd);
    NDD_VERIFY_POINTER(query_fvd);
    NDD_VERIFY_POINTER(stored_sfd);
    NDD_VERIFY_POINTER(stored_ffd);
    NDD_VERIFY_POINTER(stored_fvd);

    ndd_bool is_same = false;
    NDD_VERIFY_RESULT(ndd_sfd_compare_format(query_sfd, stored_sfd, &is_same));
    NDD_VERIFY_RESULT(ndd_ffd_compare_format(query_ffd, stored_ffd, &is_same));
    NDD_VERIFY_RESULT(ndd_fvd_compare_format(query_fvd, stored_fvd, &is_same));

    NDD_VERIFY_RESULT(ndd_fvd_header_compare_format(query_fvd->header,
						    stored_ffd->header, &is_same));
    NDD_VERIFY_RESULT(ndd_ffd_header_compare_format(query_ffd->header,
						    stored_sfd->header, &is_same));
    NDD_VERIFY_RESULT(ndd_ffs_data_compare(query_ffd->ffs, query_ffd->header,
					   query_sfd->ffs, stored_sfd->header, &is_same));

    /*  ѥ᡼ΰפå  */
    ndd_uint32 elements_per_frame = 0;
    NDD_VERIFY_RESULT(ndd_file_header_get_elements_per_frame(query_fvd->header, &elements_per_frame));

    /*  STEP0  */
    /*  Nh ׻  */
    ndd_uint32 segment_skip_1 = 0;
    NDD_IFOK{
	if(query_sfd->header->segment_size > query_sfd->header->minimum_segment_size){
	    NDD_SET_ERROR(NDD_ERROR_MINIMUM_SEGMENT_SIZE_TOO_SMALL);
	}else{
	    segment_skip_1 = query_sfd->header->minimum_segment_size - query_sfd->header->segment_size + 1;
	}
    }
    ndd_uint32 segment_skip_2 = 0;

    ndd_float8 threshold_1 = 0, threshold_2 = 0, threshold_3 = 0;  /*  ºݤͤΣ  */
    NDD_IFOK{
	threshold_1 = query_sfd->header->sfd_search_threshold * query_sfd->header->sfd_search_threshold;
	threshold_2 = query_ffd->header->ffd_search_threshold * query_ffd->header->ffd_search_threshold;
	threshold_3 = query_fvd->header->fvd_search_threshold * query_fvd->header->fvd_search_threshold;
    }

    /*  FVD ǡѥХåե  */
    ndd_float8 *frame_buffer_1 = NULL;
    ndd_float8 *frame_buffer_2 = NULL;
    NDD_VERIFY_RESULT(ndd_malloc(&frame_buffer_1, sizeof(ndd_float8) * elements_per_frame));
    NDD_VERIFY_RESULT(ndd_malloc(&frame_buffer_2, sizeof(ndd_float8) * elements_per_frame));

    ndd_uint32   answer_count = 0;
    ndd_answer **answers = NULL;

    ndd_uint32 count0 = 0, count1 = 0, count2 = 0, count3 = 0;
    ndd_uint32 segment_index_1;
    ndd_uint32 segment_index_2;
    NDD_FOR(segment_index_1 = 0; segment_index_1 < query_sfd->header->segment_count; segment_index_1++){
	NDD_IFOK{
	    if(callback != NULL){
		NDD_VERIFY_RESULT(callback("search", "", segment_index_1, query_sfd->header->segment_count, user_data));
	    }
	}

	ndd_uint32 segment_index_2_start = (query_sfd == stored_sfd) ? segment_index_1 + segment_skip_2 : 0;
	NDD_FOR(segment_index_2 = segment_index_2_start; segment_index_2 < stored_sfd->header->segment_count; segment_index_2 += segment_skip_1){
	    ndd_float8 distance = 0;
	    count0++;

	    /*  SFD ˤ  */
	    NDD_IFOK{
		distance = calculate_distance(query_sfd->sfd->segment_vector[segment_index_1],
					      stored_sfd->sfd->segment_vector[segment_index_2],
					      query_sfd->header->segment_dimension, threshold_1);
	    }
	    NDD_IFOK{
		if(distance > threshold_1){
		    continue;
		}
		count1++;
	    }

	    /*  FFD ˤ  */
	    distance = 0;
	    ndd_uint32 frame_index;
	    NDD_FOR(frame_index = 0; distance < threshold_2 && frame_index < query_sfd->header->segment_size; frame_index++){
		ndd_float8 d = 0;
		NDD_IFOK{
		    d = calculate_distance(query_ffd->ffd->frame_vector[segment_index_1 + frame_index],
					   stored_ffd->ffd->frame_vector[segment_index_2 + frame_index],
					   query_ffd->header->frame_dimension, threshold_2);
		}

		distance += d;
	    }
	    NDD_IFOK{
		if(distance > threshold_2){
		    continue;
		}
		count2++;
	    }

	    /*  FVD ˤ  */
	    distance = 0;
	    NDD_FOR(frame_index = 0; distance < threshold_3 && frame_index < query_sfd->header->segment_size; frame_index++){

		NDD_VERIFY_RESULT(ndd_fvd_read_frame_data(query_fvd, segment_index_1 + frame_index,
							  frame_buffer_1, elements_per_frame));
		NDD_VERIFY_RESULT(ndd_fvd_read_frame_data(stored_fvd, segment_index_2 + frame_index,
							  frame_buffer_2, elements_per_frame));
		ndd_float8 d = 0;
		NDD_IFOK{
		    d = calculate_distance(frame_buffer_1, frame_buffer_2, elements_per_frame, threshold_3);
		}

		distance += d;
	    }
	    NDD_IFOK{
		if(distance > threshold_3){
		    continue;
		}
		count3++;
	    }

	    /*  ޥåս  */
	    ndd_answer *new_answer = NULL;
	    NDD_VERIFY_RESULT(ndd_answer_create(&new_answer));
	    NDD_IFOK{
		new_answer->frame_offset_1 = segment_index_1;
		new_answer->frame_offset_2 = segment_index_2;
		new_answer->frame_count    = query_sfd->header->segment_size;
		new_answer->segment_count  = 1;
		new_answer->distance_min = new_answer->distance_max = new_answer->distance_avg = sqrt(distance);
	    }

	    NDD_VERIFY_RESULT(ndd_realloc(&answers, sizeof(ndd_answer *) * (answer_count + 1)));
	    NDD_IFOK{
		answers[answer_count++] = new_answer;
		new_answer = NULL;
	    }
	    
	    if(new_answer){ ndd_answer_destroy(&new_answer); }
	}
    }
    NDD_IFOK{
	if(callback != NULL){
	    NDD_VERIFY_RESULT(callback("search", NULL, 0, 0, user_data));
	}
    }

    /*  merge ʳ  */
    ndd_uint32 i;
    NDD_IFOK{
	if(search_merge_level > ndd_search_merge_level_none){
	    NDD_FOR(i = 0; i < answer_count; i++){
		ndd_answer *answer_1 = answers[i];
		ndd_uint32  frame_distance_1 = answer_1->frame_offset_1 - answer_1->frame_offset_2;

		ndd_uint32 j;
		NDD_FOR(j = i + 1; j < answer_count; ){
		    ndd_answer *answer_2 = answers[j];
		    ndd_uint32  frame_distance_2 = answer_2->frame_offset_1 - answer_2->frame_offset_2;

		    if(frame_distance_1 == frame_distance_2 &&
		       answer_1->frame_offset_1 + answer_1->frame_count > answer_2->frame_offset_1){

			if(answer_1->distance_min > answer_2->distance_min){
			    answer_1->distance_min = answer_2->distance_min;
			}
			if(answer_1->distance_max < answer_2->distance_max){
			    answer_1->distance_max = answer_2->distance_max;
			}
			answer_1->distance_avg =
			    (answer_1->distance_avg * answer_1->segment_count + answer_2->distance_avg * answer_2->segment_count) /
			    (answer_1->segment_count + answer_2->segment_count);
			answer_1->frame_count    = answer_2->frame_offset_1 + answer_2->frame_count - answer_1->frame_offset_1;
			answer_1->segment_count += answer_2->segment_count;

			ndd_answer_destroy(&answer_2);
			memcpy(answers + j, answers + j + 1, sizeof(ndd_answer *) * (answer_count - j - 1));
			answer_count--;
		    }else{
			j++;
		    }
		}
	    }
	}
    }

    /*  merge ʳܽλζֿ򵭲  */
    ndd_uint32 answer_count_1 = answer_count;

    /*  ξüΤФ  */
    NDD_IFOK{
	if(search_merge_level >= ndd_search_merge_level_line){

	    NDD_FOR(i = 0; i < answer_count; i++){
		ndd_answer *answer = answers[i];

		ndd_uint32 j;
		/*  ˿Ф  */
		NDD_FOR(j = 0; answer->frame_offset_1 > 0 && answer->frame_offset_2 > 0; j++){

		    ndd_float8 distance = 0;
		    ndd_uint32 frame_index;
		    NDD_FOR(frame_index = 0; distance < threshold_3 && frame_index < query_sfd->header->segment_size; frame_index++){

			NDD_VERIFY_RESULT(ndd_fvd_read_frame_data(query_fvd, answer->frame_offset_1 - 1 + frame_index,
								  frame_buffer_1, elements_per_frame));
			NDD_VERIFY_RESULT(ndd_fvd_read_frame_data(stored_fvd, answer->frame_offset_2 - 1 + frame_index,
								  frame_buffer_2, elements_per_frame));
			ndd_float8 d = 0;
			NDD_IFOK{
			    d = calculate_distance(frame_buffer_1, frame_buffer_2, elements_per_frame, threshold_3);
			}

			distance += d;
		    }

		    if(distance > threshold_3){
			break;
		    }
		    answer->frame_offset_1--;
		    answer->frame_offset_2--;
		    answer->frame_count++;
		}

		/*  ˿Ф  */
		NDD_FOR(j = 0;
			answer->frame_offset_1 + answer->frame_count + 1 < query_fvd->header->frame_offset + query_fvd->header->frame_count &&
			answer->frame_offset_2 + answer->frame_count + 1 < stored_fvd->header->frame_offset + stored_fvd->header->frame_count;
			j++){

		    ndd_float8 distance = 0;
		    ndd_uint32 frame_index;
		    NDD_FOR(frame_index = 0; distance < threshold_3 && frame_index < query_sfd->header->segment_size; frame_index++){

			NDD_VERIFY_RESULT(ndd_fvd_read_frame_data(query_fvd, answer->frame_offset_1 + answer->frame_count - query_sfd->header->segment_size + frame_index + 1,
								  frame_buffer_1, elements_per_frame));
			NDD_VERIFY_RESULT(ndd_fvd_read_frame_data(stored_fvd, answer->frame_offset_2 + answer->frame_count - query_sfd->header->segment_size + frame_index + 1,
								  frame_buffer_2, elements_per_frame));
			ndd_float8 d = 0;
			NDD_IFOK{
			    d = calculate_distance(frame_buffer_1, frame_buffer_2, elements_per_frame, threshold_3);
			}

			distance += d;
		    }

		    if(distance > threshold_3){
			break;
		    }
		    answer->frame_count++;
		}
	    }
	}
    }

    NDD_IFOK{
	if(search_merge_level >= ndd_search_merge_level_square){
	    /*  Ĺ˥  */
	    qsort(answers, answer_count, sizeof(ndd_answer *), compare_answer_duration);

	    /*  merge ʳ  */
	    NDD_FOR(i = 0; i < answer_count; i++){
		ndd_answer *answer_1 = answers[i];

		ndd_uint32 j;
		NDD_FOR(j = i + 1; j < answer_count; ){
		    ndd_answer *answer_2 = answers[j];

		    if(answer_1->frame_offset_1 <= answer_2->frame_offset_1 &&
		       answer_1->frame_offset_2 <= answer_2->frame_offset_2 &&
		       (answer_1->frame_offset_1 + answer_1->frame_count) >= (answer_2->frame_offset_1 + answer_2->frame_count) &&
		       (answer_1->frame_offset_2 + answer_1->frame_count) >= (answer_2->frame_offset_2 + answer_2->frame_count)){
			ndd_answer_destroy(&answer_2);
			memcpy(answers + j, answers + j + 1, sizeof(ndd_answer *) * (answer_count - j - 1));
			answer_count--;
		    }else{
			j++;
		    }
		}
	    }
	}
    }

    /*  distance ˥  */
    qsort(answers, answer_count, sizeof(ndd_answer *), compare_answer_distance);

    /*  ɽ  */
    ndd_uint32 c = 0;
    NDD_FOR(i = 0; i < answer_count; i++){
	ndd_answer *answer = answers[i];
	if(answer == NULL){
	    continue;
	}

	fprintf(stdout, "[%lu] distance = Avg:%f  Min:%f  Max:%f\n",
		++c,
		answer->distance_avg, 
		answer->distance_min,
		answer->distance_max);
	fprintf(stdout, "    ");
	print_time(answer->frame_offset_1                         / query_sfd->header->segments_per_second, stdout);
	fprintf(stdout, " - ");
	print_time((answer->frame_offset_1 + answer->frame_count) / query_sfd->header->segments_per_second, stdout);
	fprintf(stdout, " (%.3fsecs)\n",
		answer->frame_count                            / query_sfd->header->segments_per_second);

	fprintf(stdout, "    ");
	print_time(answer->frame_offset_2                         / query_sfd->header->segments_per_second, stdout);
	fprintf(stdout, " - ");
	print_time((answer->frame_offset_2 + answer->frame_count) / query_sfd->header->segments_per_second, stdout);
	fprintf(stdout, " (%.3fsecs)\n",
		answer->frame_count                            / query_sfd->header->segments_per_second);

	fprintf(stdout, "\n");
    }

    fprintf(stdout, "Total matchings:     %10lu\n", count0);
    fprintf(stdout, "Passed SFD matching: %10lu\n", count1);
    fprintf(stdout, "Passed FFD matching: %10lu\n", count2);
    fprintf(stdout, "Passed FVD matching: %10lu\n", count3);
    fprintf(stdout, "After merge 1:       %10lu\n", answer_count_1);
    fprintf(stdout, "After merge 2:       %10lu\n", answer_count);

    NDD_RETURN;
}

/*  end of file  */
